import pickle
from typing import Dict, Any, Optional, Type

from ..data import GraphDataset
from .task_utils import task_name_to_dataset_class


# This should rightfully live in model_utils.py, but that would lead to a circular import...
def get_model_file_path(model_path: str, target_suffix: str):
    assert target_suffix in ("hdf5", "pkl")
    if model_path.endswith(".hdf5"):
        return model_path[:-4] + target_suffix
    elif model_path.endswith(".pkl"):
        return model_path[:-3] + target_suffix
    else:
        raise ValueError(
            f"Model path has to end in hdf5/pkl, which is not the case for {model_path}!"
        )


def load_dataset_for_prediction(trained_model_file: str):
    with open(get_model_file_path(trained_model_file, "pkl"), "rb") as in_file:
        data_to_load = pickle.load(in_file)
    dataset_class: Type[GraphDataset] = data_to_load["dataset_class"]

    return dataset_class(
        params=data_to_load.get("dataset_params", {}),
        metadata=data_to_load.get("dataset_metadata", {}),
    )


def get_dataset(
    task_name: Optional[str],
    dataset_cls: Optional[Type[GraphDataset]],
    dataset_model_optimised_default_hyperparameters: Dict[str, Any],
    loaded_data_hyperparameters: Dict[str, Any],
    cli_data_hyperparameter_overrides: Dict[str, Any],
    loaded_metadata: Dict[str, Any],
) -> GraphDataset:
    if not dataset_cls:
        (
            dataset_cls,
            dataset_default_hyperparameter_overrides,
        ) = task_name_to_dataset_class(task_name)
        dataset_params = dataset_cls.get_default_hyperparameters()
        print(f" Dataset default parameters: {dataset_params}")
        dataset_params.update(dataset_default_hyperparameter_overrides)
        if len(dataset_default_hyperparameter_overrides):
            print(
                f"  Dataset parameters overridden by task defaults: {dataset_default_hyperparameter_overrides}"
            )
        dataset_params.update(dataset_model_optimised_default_hyperparameters)
        if len(dataset_default_hyperparameter_overrides):
            print(
                f"  Dataset parameters overridden by task/model defaults: {dataset_model_optimised_default_hyperparameters}"
            )
    else:
        dataset_params = loaded_data_hyperparameters
    dataset_params.update(cli_data_hyperparameter_overrides)
    if len(cli_data_hyperparameter_overrides):
        print(
            f"  Dataset parameters overridden from CLI: {cli_data_hyperparameter_overrides}"
        )
    if len(loaded_metadata):
        print("  WARNING: Dataset metadata loaded from disk, not calculated from data.")
    return dataset_cls(dataset_params, loaded_metadata)



"""Graph neural network layer using MLPs to compute edge messages."""
from typing import Dict, List, Any

import tensorflow as tf
from dpu_utils.tf2utils import MLP

from .message_passing import MessagePassing, MessagePassingInput, register_message_passing_implementation
from tf2_gnn.utils.constants import SMALL_NUMBER


@register_message_passing_implementation
class GNN_Edge_MLP(MessagePassing):
    """Compute new graph states by neural message passing using an edge MLP.
    For this, we assume existing node states h^t_v and a list of per-edge-type adjacency
    matrices A_\ell.

    We compute new states as follows:
        h^{t+1}_v := \sum_\ell
                     \sum_{(u, v) \in A_\ell}
                        \sigma(1/c_{v,\ell} * MLP(h^t_u || h^t_v))
        c_{\v,\ell} is usually 1 (but could also be the number of incoming edges).
    The learnable parameters of this are the W_\ell, F_{\ell,\alpha}, F_{\ell,\beta} \in R^{D, D}.

    We use the following abbreviations in shape descriptions:
    * V: number of nodes
    * L: number of different edge types
    * E: number of edges of a given edge type
    * D: input node representation dimension
    * H: output node representation dimension (set as hidden_dim)

    >>> node_embeddings = tf.random.normal(shape=(5, 3))
    >>> adjacency_lists = (
    ...    tf.constant([[0, 1], [2, 4], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[2, 3], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[3, 1]], dtype=tf.int32),
    ... )
    ...
    >>> params = GNN_Edge_MLP.get_default_hyperparameters()
    >>> params["hidden_dim"] = 12
    >>> layer = GNN_Edge_MLP(params)
    >>> output = layer(MessagePassingInput(node_embeddings, adjacency_lists))
    >>> print(output)
    tf.Tensor(..., shape=(5, 12), dtype=float32)
    """

    @classmethod
    def get_default_hyperparameters(cls):
        these_hypers = {
            "use_target_state_as_input": True,
            "normalize_by_num_incoming": False,
            "num_edge_MLP_hidden_layers": 1,
        }
        mp_hypers = super().get_default_hyperparameters()
        mp_hypers.update(these_hypers)
        return mp_hypers

    def __init__(self, params: Dict[str, Any], **kwargs):
        super().__init__(params, **kwargs)
        self._use_target_state_as_input = params["use_target_state_as_input"]
        self._normalize_by_num_incoming = params["normalize_by_num_incoming"]
        self._num_edge_MLP_hidden_layers = params["num_edge_MLP_hidden_layers"]
        self._edge_type_mlps: List[tf.keras.layers.Layer] = []

    def build(self, input_shapes: MessagePassingInput):
        node_embedding_shapes = input_shapes.node_embeddings
        adjacency_list_shapes = input_shapes.adjacency_lists
        num_edge_types = len(adjacency_list_shapes)

        if self._use_target_state_as_input:
            edge_layer_input_size = 2 * node_embedding_shapes[-1]
        else:
            edge_layer_input_size = node_embedding_shapes[-1]

        for i in range(num_edge_types):
            with tf.name_scope(f"edge_type_{i}"):
                mlp = MLP(
                    out_size=self._hidden_dim, hidden_layers=self._num_edge_MLP_hidden_layers
                )
                mlp.build(tf.TensorShape((None, edge_layer_input_size)))
            self._edge_type_mlps.append(mlp)

        super().build(input_shapes)

    def _message_function(
        self,
        edge_source_states: tf.Tensor,
        edge_target_states: tf.Tensor,
        num_incoming_to_node_per_message: tf.Tensor,
        edge_type_idx: int,
        training: bool,
    ) -> tf.Tensor:
        if self._use_target_state_as_input:
            edge_mlp_inputs = tf.concat(
                [edge_source_states, edge_target_states], axis=1
            )  # Shape [E, 2*D]
        else:
            edge_mlp_inputs = edge_source_states

        # Actually do the message calculation:
        messages = self._edge_type_mlps[edge_type_idx](edge_mlp_inputs, training)

        if self._normalize_by_num_incoming:
            messages = (
                tf.expand_dims(1.0 / (num_incoming_to_node_per_message + SMALL_NUMBER), axis=-1)
                * messages
            )
        return messages


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""Graph convolution layers."""
from typing import Dict, List, Any

import tensorflow as tf

from .gnn_edge_mlp import GNN_Edge_MLP
from .message_passing import MessagePassing, MessagePassingInput, register_message_passing_implementation
from tf2_gnn.utils.constants import SMALL_NUMBER


@register_message_passing_implementation
class RGCN(GNN_Edge_MLP):
    """Compute new graph states by neural message passing.
    This implements the R-GCN model:
    (Schlichtkrull et al., https://arxiv.org/pdf/1703.06103.pdf)
    for the case of few relations / edge types, i.e., we do not use the
    dimensionality-reduction tricks from section 2.2 of that paper.
    For this, we assume existing node states h^t_v and a list of per-edge-type adjacency
    matrices A_\ell.

    We compute new states as follows:
        h^{t+1}_v := \sigma(\sum_\ell
                            \sum_{(u, v) \in A_\ell}
                               1/c_{v,\ell} * (W_\ell * h^t_u))
    c_{\v,\ell} is usually the number of \ell edges going into v.
    The learnable parameters of this are the W_\ell \in R^{D,D}.

    We use the following abbreviations in shape descriptions:
    * V: number of nodes
    * L: number of different edge types
    * E: number of edges of a given edge type
    * D: input node representation dimension
    * H: output node representation dimension (set as hidden_dim)

    >>> node_embeddings = tf.random.normal(shape=(5, 3))
    >>> adjacency_lists = (
    ...    tf.constant([[0, 1], [2, 4], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[2, 3], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[3, 1]], dtype=tf.int32),
    ... )
    ...
    >>> params = RGCN.get_default_hyperparameters()
    >>> params["hidden_dim"] = 12
    >>> layer = RGCN(params)
    >>> output = layer(MessagePassingInput(node_embeddings, adjacency_lists))
    >>> print(output)
    tf.Tensor(..., shape=(5, 12), dtype=float32)
    """

    @classmethod
    def get_default_hyperparameters(cls):
        these_hypers = {
            "use_target_state_as_input": False,
            "normalize_by_num_incoming": True,
            "num_edge_MLP_hidden_layers": 0,
        }
        mp_hypers = super().get_default_hyperparameters()
        mp_hypers.update(these_hypers)
        return mp_hypers

    def __init__(self, params: Dict[str, Any], **kwargs):
        super().__init__(params, **kwargs)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""Graph representation aggregation layer."""
from abc import abstractmethod
from typing import List, NamedTuple

import tensorflow as tf
from dpu_utils.tf2utils import MLP, get_activation_function_by_name, unsorted_segment_softmax


class NodesToGraphRepresentationInput(NamedTuple):
    """A named tuple to hold input to layers computing graph representations from nodes
    representations."""

    node_embeddings: tf.Tensor
    node_to_graph_map: tf.Tensor
    num_graphs: tf.Tensor


class NodesToGraphRepresentation(tf.keras.layers.Layer):
    """Abstract class to compute graph representations from node representations.

    Throughout we use the following abbreviations in shape descriptions:
        * V: number of nodes (across all graphs)
        * VD: node representation dimension
        * G: number of graphs
        * GD: graph representation dimension
    """

    def __init__(self, graph_representation_size: int, **kwargs):
        super().__init__(**kwargs)
        self._graph_representation_size = graph_representation_size

    @abstractmethod
    def call(self, inputs: NodesToGraphRepresentationInput, training: bool = False):
        """Call the layer.

        Args:
            inputs: A tuple containing two items:
                node_embeddings: float32 tensor of shape [V, VD], the representation of each
                    node in all graphs.
                node_to_graph_map: int32 tensor of shape [V] with values in range [0, G-1],
                    mapping each node to a graph ID.
                num_graphs: int32 scalar, specifying the number G of graphs.
            training: A bool that denotes whether we are in training mode.

        Returns:
            float32 tensor of shape [G, GD]
        """
        pass


class WeightedSumGraphRepresentation(NodesToGraphRepresentation):
    """Layer computing graph representations as weighted sum of node representations.
    The weights are either computed from the original node representations ("self-attentional")
    or by a softmax across the nodes of a graph.
    Supports splitting operation into parallely computed independent "heads" which can focus
    on different aspects.

    Throughout we use the following abbreviations in shape descriptions:
        * V: number of nodes (across all graphs)
        * VD: node representation dimension
        * G: number of graphs
        * GD: graph representation dimension
        * H: number of heads
    """

    def __init__(
        self,
        graph_representation_size: int,
        num_heads: int,
        weighting_fun: str = "softmax",  # One of {"softmax", "sigmoid"}
        scoring_mlp_layers: List[int] = [128],
        scoring_mlp_activation_fun: str = "ReLU",
        scoring_mlp_dropout_rate: float = 0.2,
        transformation_mlp_layers: List[int] = [128],
        transformation_mlp_activation_fun: str = "ReLU",
        transformation_mlp_dropout_rate: float = 0.2,
        **kwargs,
    ):
        """
        Args:
            graph_representation_size: Size of the computed graph representation.
            num_heads: Number of independent heads to use to compute weights.
            weighting_fun: "sigmoid" ([0, 1] weights for each node computed from its
                representation), "softmax" ([0, 1] weights for each node computed
                from all nodes in same graph), "average" (weight is fixed to 1/num_nodes),
                or "none" (weight is fixed to 1).
            scoring_mlp_layers: MLP layer structure for computing raw scores turned into
                weights.
            scoring_mlp_activation_fun: MLP activcation function for computing raw scores
                turned into weights.
            scoring_mlp_dropout_rate: MLP inter-layer dropout rate for computing raw scores
                turned into weights.
            transformation_mlp_layers: MLP layer structure for computing graph representations.
            transformation_mlp_activation_fun: MLP activcation function for computing graph
                representations.
            transformation_mlp_dropout_rate: MLP inter-layer dropout rate for computing graph
                representations.
        """
        super().__init__(graph_representation_size, **kwargs)
        assert (
            graph_representation_size % num_heads == 0
        ), f"Number of heads {num_heads} needs to divide final representation size {graph_representation_size}!"
        assert weighting_fun.lower() in {
            "none",
            "average",
            "softmax",
            "sigmoid",
        }, f"Weighting function {weighting_fun} unknown, {{'softmax', 'sigmoid', 'none', 'average'}} supported."

        self._num_heads = num_heads
        self._weighting_fun = weighting_fun.lower()
        self._scoring_mlp_layers = scoring_mlp_layers
        self._scoring_mlp_activation_fun = get_activation_function_by_name(
            scoring_mlp_activation_fun
        )
        self._scoring_mlp_dropout_rate = scoring_mlp_dropout_rate
        self._transformation_mlp_layers = transformation_mlp_layers
        self._transformation_mlp_activation_fun = get_activation_function_by_name(
            transformation_mlp_activation_fun
        )
        self._transformation_mlp_dropout_rate = transformation_mlp_dropout_rate

    def build(self, input_shapes: NodesToGraphRepresentationInput):
        with tf.name_scope("WeightedSumGraphRepresentation"):
            if self._weighting_fun not in ("none", "average"):
                self._scoring_mlp = MLP(
                    out_size=self._num_heads,
                    hidden_layers=self._scoring_mlp_layers,
                    activation_fun=self._scoring_mlp_activation_fun,
                    dropout_rate=self._scoring_mlp_dropout_rate,
                    name="ScoringMLP",
                )
                self._scoring_mlp.build(tf.TensorShape((None, input_shapes.node_embeddings[-1])))

            self._transformation_mlp = MLP(
                out_size=self._graph_representation_size,
                hidden_layers=self._transformation_mlp_layers,
                activation_fun=self._transformation_mlp_activation_fun,
                dropout_rate=self._transformation_mlp_dropout_rate,
                name="TransformationMLP",
            )
            self._transformation_mlp.build(tf.TensorShape((None, input_shapes.node_embeddings[-1])))

            super().build(input_shapes)

    @tf.function(
        input_signature=(
            NodesToGraphRepresentationInput(
                node_embeddings=tf.TensorSpec(shape=tf.TensorShape((None, None)), dtype=tf.float32),
                node_to_graph_map=tf.TensorSpec(shape=tf.TensorShape((None,)), dtype=tf.int32),
                num_graphs=tf.TensorSpec(shape=(), dtype=tf.int32),
            ),
            tf.TensorSpec(shape=(), dtype=tf.bool),
        )
    )
    def call(self, inputs: NodesToGraphRepresentationInput, training: bool = False):
        # (1) compute weights for each node/head pair:
        if self._weighting_fun not in ("none", "average"):
            scores = self._scoring_mlp(inputs.node_embeddings, training=training)  # Shape [V, H]
            if self._weighting_fun == "sigmoid":
                weights = tf.nn.sigmoid(scores)  # Shape [V, H]
            elif self._weighting_fun == "softmax":
                weights_per_head = []
                for head_idx in range(self._num_heads):
                    head_scores = scores[:, head_idx]  # Shape [V]
                    head_weights = unsorted_segment_softmax(
                        logits=head_scores,
                        segment_ids=inputs.node_to_graph_map,
                        num_segments=inputs.num_graphs,
                    )  # Shape [V]
                    weights_per_head.append(tf.expand_dims(head_weights, -1))
                weights = tf.concat(weights_per_head, axis=1)  # Shape [V, H]
            else:
                raise ValueError()

        # (2) compute representations for each node/head pair:
        node_reprs = self._transformation_mlp(
            inputs.node_embeddings, training=training
        )  # Shape [V, GD]
        node_reprs = tf.reshape(
            node_reprs,
            shape=(-1, self._num_heads, self._graph_representation_size // self._num_heads),
        )  # Shape [V, H, GD//H]

        # (3) if necessary, weight representations and aggregate by graph:
        if self._weighting_fun == "none":
            node_reprs = tf.reshape(
                node_reprs, shape=(-1, self._graph_representation_size)
            )  # Shape [V, GD]
            graph_reprs = tf.math.segment_sum(
                data=node_reprs, segment_ids=inputs.node_to_graph_map
            )  # Shape [G, GD]
        elif self._weighting_fun == "average":
            node_reprs = tf.reshape(
                node_reprs, shape=(-1, self._graph_representation_size)
            )  # Shape [V, GD]
            graph_reprs = tf.math.segment_mean(
                data=node_reprs, segment_ids=inputs.node_to_graph_map
            )  # Shape [G, GD]
        else:
            weights = tf.expand_dims(weights, -1)  # Shape [V, H, 1]
            weighted_node_reprs = weights * node_reprs  # Shape [V, H, GD//H]

            weighted_node_reprs = tf.reshape(
                weighted_node_reprs, shape=(-1, self._graph_representation_size)
            )  # Shape [V, GD]
            graph_reprs = tf.math.segment_sum(
                data=weighted_node_reprs, segment_ids=inputs.node_to_graph_map
            )  # Shape [G, GD]

        return graph_reprs



"""GNN Encoder class."""
from typing import Any, Dict, NamedTuple, List, Tuple, Optional

import tensorflow as tf

from tf2_gnn.utils.param_helpers import get_activation_function
from .message_passing import (
    MessagePassing,
    MessagePassingInput,
    get_message_passing_class,
)
from .graph_global_exchange import (
    GraphGlobalExchangeInput,
    GraphGlobalExchange,
    GraphGlobalMeanExchange,
    GraphGlobalGRUExchange,
    GraphGlobalMLPExchange,
)


class GNNInput(NamedTuple):
    """Input named tuple for the GNN."""

    node_features: tf.Tensor
    adjacency_lists: Tuple[tf.Tensor, ...]
    node_to_graph_map: tf.Tensor
    num_graphs: tf.Tensor


class GNN(tf.keras.layers.Layer):
    """Encode graph states using a combination of graph message passing layers and dense layers

    Example usage:
    >>> layer_input = GNNInput(
    ...     node_features = tf.random.normal(shape=(5, 3)),
    ...     adjacency_lists = (
    ...         tf.constant([[0, 1], [1, 2], [3, 4]], dtype=tf.int32),
    ...         tf.constant([[1, 2], [3, 4]], dtype=tf.int32),
    ...         tf.constant([[2, 0]], dtype=tf.int32)
    ...         ),
    ...     node_to_graph_map = tf.fill(dims=(5,), value=0),
    ...     num_graphs = 1,
    ...     )
    ...
    >>> params = GNN.get_default_hyperparameters()
    >>> params["hidden_dim"] = 12
    >>> layer = GNN(params)
    >>> output = layer(layer_input)
    >>> print(output)
    tf.Tensor(..., shape=(5, 12), dtype=float32)
    """

    @classmethod
    def get_default_hyperparameters(cls, mp_style: Optional[str] = None) -> Dict[str, Any]:
        """Get the default hyperparameter dictionary for the  class."""
        these_hypers = {
            "message_calculation_class": "rgcn",
            "initial_node_representation_activation": "tanh",
            "dense_intermediate_layer_activation": "tanh",
            "num_layers": 4,
            "dense_every_num_layers": 2,
            "residual_every_num_layers": 2,
            "use_inter_layer_layernorm": False,
            "hidden_dim": 16,
            "layer_input_dropout_rate": 0.0,
            "global_exchange_mode": "gru",  # One of "mean", "mlp", "gru"
            "global_exchange_every_num_layers": 2,
            "global_exchange_weighting_fun": "softmax",  # One of "softmax", "sigmoid"
            "global_exchange_num_heads": 4,
            "global_exchange_dropout_rate": 0.2,
        }  # type: Dict[str, Any]
        if mp_style is not None:
            these_hypers["message_calculation_class"] = mp_style
        message_passing_class = get_message_passing_class(
            these_hypers["message_calculation_class"]
        )
        message_passing_hypers = message_passing_class.get_default_hyperparameters()
        message_passing_hypers.update(these_hypers)
        return message_passing_hypers

    def __init__(self, params: Dict[str, Any]):
        """Initialise the layer."""
        super().__init__()
        self._params = params
        self._hidden_dim = params["hidden_dim"]
        self._num_layers = params["num_layers"]
        self._dense_every_num_layers = params["dense_every_num_layers"]
        self._residual_every_num_layers = params["residual_every_num_layers"]
        self._use_inter_layer_layernorm = params["use_inter_layer_layernorm"]
        self._initial_node_representation_activation_fn = get_activation_function(
            params["initial_node_representation_activation"]
        )
        self._dense_intermediate_layer_activation_fn = get_activation_function(
            params["dense_intermediate_layer_activation"]
        )
        self._message_passing_class = get_message_passing_class(
            params["message_calculation_class"]
        )

        if not params["global_exchange_mode"].lower() in {"mean", "mlp", "gru"}:
            raise ValueError(
                f"Unknown global_exchange_mode mode {params['global_exchange_mode']} - has to be one of 'mean', 'mlp', 'gru'!"
            )
        self._global_exchange_mode = params["global_exchange_mode"]
        self._global_exchange_every_num_layers = params["global_exchange_every_num_layers"]
        self._global_exchange_weighting_fun = params["global_exchange_weighting_fun"]
        self._global_exchange_num_heads = params["global_exchange_num_heads"]
        self._global_exchange_dropout_rate = params["global_exchange_dropout_rate"]

        # Layer member variables. To be filled in in the `build` method.
        self._initial_projection_layer: tf.keras.layers.Layer = None
        self._mp_layers: List[MessagePassing] = []
        self._inter_layer_layernorms: List[tf.keras.layers.Layer] = []
        self._dense_layers: Dict[str, tf.keras.layers.Layer] = {}
        self._global_exchange_layers: Dict[str, GraphGlobalExchange] = {}

    def build(self, tensor_shapes: GNNInput):
        """Build the various layers in the model.

        Args:
            tensor_shapes: A GNNInput of tensor shapes.

        Returns:
            Nothing, but initialises the layers in the model based on the tensor shapes given.
        """
        # First, we go through the input shapes and make sure that anything which might vary batch
        # to batch (number of nodes / number of edges) is set to None.
        initial_node_features_shape: tf.TensorShape = tensor_shapes.node_features
        variable_node_features_shape = tf.TensorShape((None, initial_node_features_shape[1]))
        adjacency_list_shapes = tensor_shapes.adjacency_lists
        embedded_shape = tf.TensorShape((None, self._hidden_dim))

        with tf.name_scope(f"{self._message_passing_class.__name__}_GNN"):
            # Then we construct the layers themselves:
            with tf.name_scope("gnn_initial_node_projection"):
                self._initial_projection_layer = tf.keras.layers.Dense(
                    units=self._hidden_dim,
                    use_bias=False,
                    activation=self._initial_node_representation_activation_fn,
                )
                self._initial_projection_layer.build(variable_node_features_shape)

            # Construct the graph message passing layers.
            for layer_idx in range(self._num_layers):
                with tf.name_scope(f"Layer_{layer_idx}"):
                    with tf.name_scope("MessagePassing"):
                        self._mp_layers.append(
                            self._message_passing_class(self._params)
                        )
                        self._mp_layers[-1].build(
                            MessagePassingInput(embedded_shape, adjacency_list_shapes)
                        )

                    # If required, prepare for a LayerNorm:
                    if self._use_inter_layer_layernorm:
                        with tf.name_scope(f"LayerNorm"):
                            self._inter_layer_layernorms.append(
                                tf.keras.layers.LayerNormalization()
                            )
                            self._inter_layer_layernorms[-1].build(embedded_shape)

                    # Construct the per-node dense layers.
                    if layer_idx % self._dense_every_num_layers == 0:
                        with tf.name_scope(f"Dense"):
                            self._dense_layers[str(layer_idx)] = tf.keras.layers.Dense(
                                units=self._hidden_dim,
                                use_bias=False,
                                activation=self._dense_intermediate_layer_activation_fn,
                            )
                            self._dense_layers[str(layer_idx)].build(embedded_shape)

                    if (
                        layer_idx
                        and layer_idx % self._global_exchange_every_num_layers == 0
                    ):
                        with tf.name_scope(f"Global_Exchange"):
                            if self._global_exchange_mode.lower() == "mean":
                                exchange_layer_class = GraphGlobalMeanExchange
                            elif self._global_exchange_mode.lower() == "gru":
                                exchange_layer_class = GraphGlobalGRUExchange
                            elif self._global_exchange_mode.lower() == "mlp":
                                exchange_layer_class = GraphGlobalMLPExchange
                            exchange_layer = exchange_layer_class(
                                hidden_dim=self._hidden_dim,
                                weighting_fun=self._global_exchange_weighting_fun,
                                num_heads=self._global_exchange_num_heads,
                                dropout_rate=self._global_exchange_dropout_rate,
                            )
                            exchange_layer.build(
                                GraphGlobalExchangeInput(
                                    node_embeddings=tf.TensorShape(
                                        (None, self._hidden_dim)
                                    ),
                                    node_to_graph_map=tf.TensorShape((None,)),
                                    num_graphs=tf.TensorShape(()),
                                )
                            )
                            self._global_exchange_layers[
                                str(layer_idx)
                            ] = exchange_layer

        super().build(tensor_shapes)

        # The following is needed to work around a limitation in the @tf.function annotation.
        # (See https://github.com/tensorflow/tensorflow/issues/32457 for a related issue,
        #  though there are many more).
        # Our aim is to trace the `call` function once and for all. However, as the first
        # dimension of node features and adjacency lists keeps changing between batches (with
        # the number of nodes/edges in the batch), generalisation doesn't work automatically.
        # Instead, we have to specify the input spec explicitly; but as this depends on a
        # build-time constant (the number of edges), we cannot do that by just using @tf.function.
        # Instead, we construct the TensorSpec explicitly, and then use setattr to wrap
        # our function using tf.function.
        #
        # Finally, the `return_all_representations` option changes the shape of the return values,
        # but a tf.function-traced function must return the same shape on all code paths. To
        # handle this, we let the core function _always_ return all representations (and trace
        # that for performance reasons), and then use a thin wrapper `call` function to drop
        # the unneeded return value if needed.
        internal_call_input_spec = (
            GNNInput(
                node_features=tf.TensorSpec(shape=variable_node_features_shape, dtype=tf.float32),
                adjacency_lists=tuple(
                    tf.TensorSpec(shape=(None, 2), dtype=tf.int32)
                    for _ in range(len(adjacency_list_shapes))
                ),
                node_to_graph_map=tf.TensorSpec(shape=(None,), dtype=tf.int32),
                num_graphs=tf.TensorSpec(shape=(), dtype=tf.int32),
            ),
            tf.TensorSpec(shape=(), dtype=tf.bool)
        )
        setattr(self, "_internal_call", tf.function(func=self._internal_call, input_signature=internal_call_input_spec))

    def call(self, inputs: GNNInput, training: bool = False, return_all_representations: bool = False):
        """
        Args:
            inputs: A GNNInput containing the following fields:
                node_features: float32 tensor of shape [V, D], the original representation
                    of each node in the graph.

                adjacency_lists: an tuple of tensors of shape [E, 2] which represents an adjacency
                    list for a given edge type. Concretely,
                        adjacency_list[l][k,:] == [v, u]
                    means that the k-th edge of type l connects node v to node u.

                node_to_graph_map: int32 tensor of shape [V], where node_to_graph_map[v] = i
                    means that node v belongs to graph i in the batch.

                num_graphs: int32 tensor of shape [], specifying number of graphs in batch.

            training: A bool representing whether the model is training or evaluating.

            return_all_representations: A bool indicating whether to return all (initial,
                intermediate, and final) GNN results as well.

        Returns:
            If return_all_representations is False (the default):
            A tensor of shape [V, hidden_dim], where hidden_dim was defined in the layer
            initialisation. The tensor represents the encoding of the initial node_features by the
            GNN framework.

            If return_all_representations is True:
            A pair, first element as for return_all_representations=False, second element a  list
            of Tensors of shape [V, hidden_dim], where the first element is the original GNN
            input (after a potential projection layer) and the remaining elements are the
            output of all GNN layers (without dropout, residual connections, dense layers 
            or layer norm applied).
        """
        cur_node_representations, all_node_representations = self._internal_call(inputs, training)

        if return_all_representations:
            return cur_node_representations, all_node_representations

        return cur_node_representations

    def _internal_call(self, inputs: GNNInput, training: bool = False):
        initial_node_features: tf.Tensor = inputs.node_features
        adjacency_lists = inputs.adjacency_lists
        cur_node_representations = self._initial_projection_layer(initial_node_features)

        # Layer loop.
        last_node_representations = cur_node_representations
        all_node_representations = [cur_node_representations]
        for layer_idx, mp_layer in enumerate(self._mp_layers):
            if training:
                cur_node_representations = tf.nn.dropout(
                    cur_node_representations, rate=self._params["layer_input_dropout_rate"]
                )

            # Pass residuals through:
            if layer_idx % self._residual_every_num_layers == 0:
                tmp = cur_node_representations
                if layer_idx > 0:
                    cur_node_representations += last_node_representations
                    cur_node_representations /= 2
                last_node_representations = tmp

            # Apply this message passing layer.
            cur_node_representations = mp_layer(
                MessagePassingInput(
                    node_embeddings=cur_node_representations, adjacency_lists=adjacency_lists
                ),
                training=training,
            )
            all_node_representations.append(cur_node_representations)

            if layer_idx and layer_idx % self._global_exchange_every_num_layers == 0:
                cur_node_representations = self._global_exchange_layers[str(layer_idx)](
                    GraphGlobalExchangeInput(
                        node_embeddings=cur_node_representations,
                        node_to_graph_map=inputs.node_to_graph_map,
                        num_graphs=inputs.num_graphs,
                    ),
                    training=training,
                )

            # If required, apply a LayerNorm:
            if self._use_inter_layer_layernorm:
                cur_node_representations = self._inter_layer_layernorms[layer_idx](
                    cur_node_representations
                )

            # Apply dense layer, if needed.
            if layer_idx % self._dense_every_num_layers == 0:
                cur_node_representations = self._dense_layers[str(layer_idx)](
                    cur_node_representations, training=training
                )

        return cur_node_representations, tuple(all_node_representations)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""Functions to convert from string parameters to their values."""
import tensorflow as tf

from .activation import gelu


def get_aggregation_function(aggregation_fn_name: str):
    """Convert from an aggregation function name to the function itself."""
    string_to_aggregation_fn = {
        "sum": tf.math.unsorted_segment_sum,
        "max": tf.math.unsorted_segment_max,
        "mean": tf.math.unsorted_segment_mean,
        "sqrt_n": tf.math.unsorted_segment_sqrt_n,
    }
    aggregation_fn = string_to_aggregation_fn.get(aggregation_fn_name)
    if aggregation_fn is None:
        raise ValueError(f"Unknown aggregation function: {aggregation_fn_name}")
    return aggregation_fn


def get_activation_function(activation_fn_name: str):
    """Convert from an activation function name to the function itself."""
    if activation_fn_name is None:
        return None
    activation_fn_name = activation_fn_name.lower()

    string_to_activation_fn = {
        "linear": None,
        "tanh": tf.nn.tanh,
        "relu": tf.nn.relu,
        "leaky_relu": tf.nn.leaky_relu,
        "elu": tf.nn.elu,
        "selu": tf.nn.selu,
        "gelu": gelu,
    }
    activation_fn = string_to_activation_fn.get(activation_fn_name)
    if activation_fn is None:
        raise ValueError(f"Unknown activation function: {activation_fn_name}")
    return activation_fn



"""General task for graph binary classification."""
from typing import Any, Dict, List, Tuple, Optional

import numpy as np
import tensorflow as tf

from tf2_gnn.data import GraphDataset
from tf2_gnn.models import GraphTaskModel
from tf2_gnn.layers import WeightedSumGraphRepresentation, NodesToGraphRepresentationInput


class GraphBinaryClassificationTask(GraphTaskModel):
    @classmethod
    def get_default_hyperparameters(cls, mp_style: Optional[str] = None) -> Dict[str, Any]:
        super_params = super().get_default_hyperparameters(mp_style)
        these_hypers: Dict[str, Any] = {
            "graph_aggregation_num_heads": 16,
            "graph_aggregation_hidden_layers": [128],
            "graph_aggregation_dropout_rate": 0.2,
        }
        super_params.update(these_hypers)
        return super_params

    def __init__(self, params: Dict[str, Any], dataset: GraphDataset, name: str = None):
        super().__init__(params, dataset=dataset, name=name)
        self._node_to_graph_aggregation = None

    def build(self, input_shapes):
        with tf.name_scope(self.__class__.__name__):
            self._node_to_graph_repr_layer = WeightedSumGraphRepresentation(
                graph_representation_size=self._params["graph_aggregation_num_heads"],
                num_heads=self._params["graph_aggregation_num_heads"],
                scoring_mlp_layers=self._params["graph_aggregation_hidden_layers"],
                scoring_mlp_dropout_rate=self._params["graph_aggregation_dropout_rate"],
                transformation_mlp_layers=self._params["graph_aggregation_hidden_layers"],
                transformation_mlp_dropout_rate=self._params["graph_aggregation_dropout_rate"],
            )
            self._node_to_graph_repr_layer.build(
                NodesToGraphRepresentationInput(
                    node_embeddings=tf.TensorShape(
                        (None, input_shapes["node_features"][-1] + self._params["gnn_hidden_dim"])
                    ),
                    node_to_graph_map=tf.TensorShape((None)),
                    num_graphs=tf.TensorShape(()),
                )
            )

            self._graph_repr_to_classification_layer = tf.keras.layers.Dense(
                units=1, activation=tf.nn.sigmoid, use_bias=True
            )
            self._graph_repr_to_classification_layer.build(
                tf.TensorShape((None, self._params["graph_aggregation_num_heads"]))
            )

        super().build(input_shapes)

    def compute_task_output(
        self,
        batch_features: Dict[str, tf.Tensor],
        final_node_representations: tf.Tensor,
        training: bool,
    ) -> Any:
        per_graph_results = self._node_to_graph_repr_layer(
            NodesToGraphRepresentationInput(
                node_embeddings=tf.concat(
                    [batch_features["node_features"], final_node_representations], axis=-1
                ),
                node_to_graph_map=batch_features["node_to_graph_map"],
                num_graphs=batch_features["num_graphs_in_batch"],
            )
        )  # Shape [G, graph_aggregation_num_heads]
        per_graph_results = self._graph_repr_to_classification_layer(
            per_graph_results
        )  # Shape [G, 1]

        return tf.squeeze(per_graph_results, axis=-1)

    def compute_task_metrics(
        self,
        batch_features: Dict[str, tf.Tensor],
        task_output: Any,
        batch_labels: Dict[str, tf.Tensor],
    ) -> Dict[str, tf.Tensor]:
        ce = tf.reduce_mean(
            tf.keras.losses.binary_crossentropy(
                y_true=batch_labels["target_value"], y_pred=task_output, from_logits=False
            )
        )
        num_correct = tf.reduce_sum(
            tf.cast(
                tf.math.equal(batch_labels["target_value"], tf.math.round(task_output)), tf.int32
            )
        )
        num_graphs = tf.cast(batch_features["num_graphs_in_batch"], tf.float32)
        return {
            "loss": ce,
            "batch_acc": tf.cast(num_correct, tf.float32) / num_graphs,
            "num_correct": num_correct,
            "num_graphs": num_graphs,
        }

    def compute_epoch_metrics(self, task_results: List[Any]) -> Tuple[float, str]:
        total_num_graphs = np.sum(
            batch_task_result["num_graphs"] for batch_task_result in task_results
        )
        total_num_correct = np.sum(
            batch_task_result["num_correct"] for batch_task_result in task_results
        )
        epoch_acc = tf.cast(total_num_correct, tf.float32) / total_num_graphs
        return -epoch_acc.numpy(), f"Accuracy = {epoch_acc.numpy():.3f}"



"""General dataset class for datasets stored as JSONLines files."""
import logging
from typing import Any, Dict, Iterator, List, Optional, Tuple, Set

import numpy as np
from dpu_utils.utils import RichPath

from .graph_dataset import DataFold, GraphDataset, GraphSampleType, GraphSample

logger = logging.getLogger(__name__)


class JsonLGraphDataset(GraphDataset[GraphSampleType]):
    """
    General class representing pre-split datasets in JSONLines format.
    Concretely, this class expects the following:
    * In the data directory, files "train.jsonl.gz", "valid.jsonl.gz" and
      "test.jsonl.gz" are used to store the train/valid/test datasets.
    * Each of the files is gzipped text file in which each line is a valid
      JSON dictionary with a "graph" key, which in turn points to a
      dictionary with keys
       - "node_features" (list of numerical initial node labels)
       - "adjacency_lists" (list of list of directed edge pairs)
    """

    @classmethod
    def get_default_hyperparameters(cls) -> Dict[str, Any]:
        super_hypers = super().get_default_hyperparameters()
        this_hypers = {
            "num_fwd_edge_types": 3,
            "add_self_loop_edges": True,
            "tie_fwd_bkwd_edges": True,
        }
        super_hypers.update(this_hypers)
        return super_hypers

    def __init__(
        self, params: Dict[str, Any], metadata: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(params, metadata=metadata)
        self._params = params
        self._num_fwd_edge_types = params["num_fwd_edge_types"]
        if params["tie_fwd_bkwd_edges"]:
            self._num_edge_types = self._num_fwd_edge_types
        else:
            self._num_edge_types = 2 * self._num_fwd_edge_types
        self._num_edge_types += int(params["add_self_loop_edges"])

        self._node_feature_shape: Optional[Tuple[int]] = None
        self._loaded_data: Dict[DataFold, List[GraphSampleType]] = {}

    @property
    def num_edge_types(self) -> int:
        return self._num_edge_types

    @property
    def node_feature_shape(self) -> Tuple:
        """Return the shape of the node features."""
        if self._node_feature_shape is None:
            some_data_fold = next(iter(self._loaded_data.values()))
            self._node_feature_shape = (len(some_data_fold[0].node_features[0]),)
        return self._node_feature_shape

    def load_metadata(self, path: RichPath) -> None:
        """Load the metadata for a dataset (such as vocabularies, names of properties, ...)
        from a path on disk.

        Note: Implementors needing to act on metadata information before loading any actual data
        should override this method.
        """
        if self.metadata == {}:
            metadata_path = path.join("metadata.pkl.gz")
            if metadata_path.exists():
                logger.info(f"Loading metadata from {metadata_path}")
                self._metadata = metadata_path.read_by_file_suffix()
        else:
            logger.warning("Using metadata passed to constructor, not metadata stored with data.")

    def load_data(self, path: RichPath, folds_to_load: Optional[Set[DataFold]] = None) -> None:
        """Load the data from disk."""
        logger.info(f"Starting to load data from {path}.")
        self.load_metadata(path)

        # If we haven't defined what folds to load, load all:
        if folds_to_load is None:
            folds_to_load = {DataFold.TRAIN, DataFold.VALIDATION, DataFold.TEST}

        if DataFold.TRAIN in folds_to_load:
            self._loaded_data[DataFold.TRAIN] = self.__load_data(path.join("train.jsonl.gz"))
            logger.debug("Done loading training data.")
        if DataFold.VALIDATION in folds_to_load:
            self._loaded_data[DataFold.VALIDATION] = self.__load_data(path.join("valid.jsonl.gz"))
            logger.debug("Done loading validation data.")
        if DataFold.TEST in folds_to_load:
            self._loaded_data[DataFold.TEST] = self.__load_data(path.join("test.jsonl.gz"))
            logger.debug("Done loading test data.")

    def load_data_from_list(
        self, datapoints: List[Dict[str, Any]], target_fold: DataFold = DataFold.TEST
    ):
        if target_fold not in self._loaded_data:
            self._loaded_data[target_fold] = []
        for datapoint in datapoints:
            self._loaded_data[target_fold].append(self._process_raw_datapoint(datapoint))

    def __load_data(self, data_file: RichPath) -> List[GraphSampleType]:
        return [
            self._process_raw_datapoint(datapoint) for datapoint in data_file.read_by_file_suffix()
        ]

    def _process_raw_datapoint(self, datapoint: Dict[str, Any]) -> GraphSampleType:
        node_features = datapoint["graph"]["node_features"]
        type_to_adj_list, type_to_num_incoming_edges = self._process_raw_adjacency_lists(
            raw_adjacency_lists=datapoint["graph"]["adjacency_lists"],
            num_nodes=len(node_features),
        )

        return GraphSample(
            adjacency_lists=type_to_adj_list,
            type_to_node_to_num_inedges=type_to_num_incoming_edges,
            node_features=node_features,
        )

    def _process_raw_adjacency_lists(
        self, raw_adjacency_lists: List[List[Tuple]], num_nodes: int
    ) -> Tuple[List, np.ndarray]:
        type_to_adj_list = [
            [] for _ in range(self._num_fwd_edge_types + int(self.params["add_self_loop_edges"]))
        ]  # type: List[List[Tuple[int, int]]]
        type_to_num_incoming_edges = np.zeros(shape=(self.num_edge_types, num_nodes))
        for raw_edge_type, edges in enumerate(raw_adjacency_lists):
            if self.params["add_self_loop_edges"]:
                fwd_edge_type = raw_edge_type + 1  # 0 will be the self-loop type
            else:
                fwd_edge_type = raw_edge_type  # Make edges start from 0
            for src, dest in edges:
                type_to_adj_list[fwd_edge_type].append((src, dest))
                type_to_num_incoming_edges[fwd_edge_type, dest] += 1
                if self.params["tie_fwd_bkwd_edges"]:
                    type_to_adj_list[fwd_edge_type].append((dest, src))
                    type_to_num_incoming_edges[fwd_edge_type, src] += 1

        if self.params["add_self_loop_edges"]:
            # Add self-loop edges (idx 0, which isn't used in the data):
            for node in range(num_nodes):
                type_to_num_incoming_edges[0, node] = 1
                type_to_adj_list[0].append((node, node))

        # Add backward edges as an additional edge type that goes backwards:
        if not (self.params["tie_fwd_bkwd_edges"]):
            # for (edge_type, adj_list) in enumerate(type_to_adj_list):
            num_edge_types_in_adj_lists = len(type_to_adj_list)
            for edge_type in range(num_edge_types_in_adj_lists):
                adj_list = type_to_adj_list[edge_type]
                # Don't add self loops again!
                if edge_type == 0 and self.params["add_self_loop_edges"]:
                    continue
                bkwd_edge_type = len(type_to_adj_list)
                type_to_adj_list.append([(y, x) for (x, y) in adj_list])
                for (x, y) in adj_list:
                    type_to_num_incoming_edges[bkwd_edge_type][y] += 1

        # Convert the adjacency lists to numpy arrays.
        type_to_adj_list = [
            np.array(adj_list, dtype=np.int32)
            if len(adj_list) > 0
            else np.zeros(shape=(0, 2), dtype=np.int32)
            for adj_list in type_to_adj_list
        ]
        return type_to_adj_list, type_to_num_incoming_edges

    def _graph_iterator(self, data_fold: DataFold) -> Iterator[GraphSampleType]:
        if data_fold == DataFold.TRAIN:
            np.random.shuffle(self._loaded_data[data_fold])
        return iter(self._loaded_data[data_fold])



from abc import abstractmethod
from enum import Enum
from typing import Any, Dict, List, Iterator, Tuple, TypeVar, Generic, NamedTuple, Set, Optional

import numpy as np
import tensorflow as tf
from dpu_utils.utils import RichPath, DoubleBufferedIterator


class DataFold(Enum):
    TRAIN = 0
    VALIDATION = 1
    TEST = 2


class GraphBatchTFDataDescription(NamedTuple):
    batch_features_types: Dict[str, Any]
    batch_features_shapes: Dict[str, Any]
    batch_labels_types: Dict[str, Any]
    batch_labels_shapes: Dict[str, Any]


class GraphSample(object):
    """Data structure holding information for a single graph."""

    def __init__(
        self,
        adjacency_lists: List[np.ndarray],
        type_to_node_to_num_inedges: np.ndarray,
        node_features: np.ndarray,
    ):
        super().__init__()
        self._adjacency_lists = adjacency_lists
        self._type_to_node_to_num_inedges = type_to_node_to_num_inedges
        self._node_features = node_features

    @property
    def adjacency_lists(self) -> List[np.ndarray]:
        """Adjacency information by edge type as list of ndarrays of shape [E, 2]"""
        return self._adjacency_lists

    @property
    def type_to_node_to_num_inedges(self) -> np.ndarray:
        """Number of incoming edge by edge type as ndarray of shape [V]"""
        return self._type_to_node_to_num_inedges

    @property
    def node_features(self) -> np.ndarray:
        """Initial node features as ndarray of shape [V, ...]"""
        return self._node_features


GraphSampleType = TypeVar("GraphSampleType", bound=GraphSample)


class GraphDataset(Generic[GraphSampleType]):
    """Abstract class turning graph datasets into iterators providing minibatches of graph
    samples.
    This class implements the core logic for batching of graphs into one huge graph with
    disconnected components, and provides hooks to insert additional data.

    To use this for your own data type, graph samples need to be stored as datapoints
    extending the GraphSample class.
    Besides implementing the abstract methods, implementors should override _new_batch,
    _add_graph_to_batch, and _finalise_batch to add dataset-specific things (such as
    labels for the graphs or nodes) to the generated minibatches.
    """

    @classmethod
    def get_default_hyperparameters(cls) -> Dict[str, Any]:
        return {
            "max_nodes_per_batch": 10000,
        }

    def __init__(self, params: Dict[str, Any], metadata: Optional[Dict[str, Any]] = None):
        self._params = params
        self._metadata = metadata if metadata is not None else {}

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def params(self) -> Dict[str, Any]:
        return self._params

    @property
    def metadata(self) -> Dict[str, Any]:
        return self._metadata

    @property
    @abstractmethod
    def num_edge_types(self) -> int:
        pass

    @property
    @abstractmethod
    def node_feature_shape(self) -> Tuple:
        pass

    @abstractmethod
    def load_data(self, path: RichPath, folds_to_load: Optional[Set[DataFold]] = None) -> None:
        pass

    @abstractmethod
    def load_data_from_list(
        self, datapoints: List[Dict[str, Any]], target_fold: DataFold = DataFold.TEST
    ):
        pass

    @abstractmethod
    def _graph_iterator(self, data_fold: DataFold) -> Iterator[GraphSampleType]:
        """Returns iterator over data in specified fold of the dataset.
        Note: The iterator is expected to shuffle training data on every call.
        """
        pass

    def graph_batch_iterator(
        self, data_fold: DataFold
    ) -> Iterator[Tuple[Dict[str, Any], Dict[str, Any]]]:
        """Get the graph batch iterator.

        Graph are batched into a single graph with multiple disconnected components.
        The "batch" is the full graph, and each disconnected component is one of the
        sample graphs that make up the batch.

        We use the following abbreviations in shape descriptions:
        * V: number of nodes
        * D: state dimension
        * L: number of different edge types
        * E: number of edges of a given edge type
        * G: number of graphs in the batch

        Returns:
            An iterator which outputs a single minibatch as a pair of dictionaries
            representing features (i.e., the graphs) and labels (e.g., regression
            targets).

            The features dictionary has at least these three entries (which are
            consumed by the "general" infrastructure):
                * node_features: a numpy float32 array of shape [V, ...] representing the
                    features for each node in the graph. Usually a one-hot-encoding of
                    the atom type, tokens in label, etc.
                * node_to_graph_map: a numpy int32 array of shape [V] that represents
                    which graph in the batch each node comes from.
                * adjacency_lists: a list of length L of numpy int32 arrays of shape
                    [E, 2] which represents an adjacency list for a given edge type.
                    Concretely, adjacency_lists[l][k,:] == [v, u]
                    means that the k-th edge of type l connects node v to node u.
        """
        graph_sample_iterator = self._graph_iterator(data_fold)
        for graph_batch in self.graph_batch_iterator_from_graph_iterator(graph_sample_iterator):
            yield graph_batch

    def graph_batch_iterator_from_graph_iterator(
        self, graph_sample_iterator: Iterator[GraphSampleType]
    ) -> Iterator[Tuple[Dict[str, Any], Dict[str, Any]]]:
        raw_batch = self._new_batch()
        for graph_sample in graph_sample_iterator:
            num_nodes_in_graph = len(graph_sample.node_features)

            # Yield the batch if adding the current graph_sample would make it too big.
            if self._batch_would_be_too_full(raw_batch, graph_sample):
                yield self._finalise_batch(raw_batch)
                raw_batch = self._new_batch()

            # Extend the bath with the current graph_sample.
            self._add_graph_to_batch(raw_batch, graph_sample)

            # Increment counters.
            raw_batch["num_graphs_in_batch"] += 1
            raw_batch["num_nodes_in_batch"] += num_nodes_in_graph

        yield self._finalise_batch(raw_batch)

    def _batch_would_be_too_full(
        self, raw_batch: Dict[str, Any], graph_sample: GraphSampleType
    ) -> bool:
        """Return whether the current raw batch would be too full if graph_sample was added."""
        num_nodes_in_graph = len(graph_sample.node_features)
        return (
            raw_batch["num_nodes_in_batch"] + num_nodes_in_graph
            > self._params["max_nodes_per_batch"]
        )

    def _new_batch(self) -> Dict[str, Any]:
        """Return a dictionary suitable for collecting data for a fresh minibatch."""
        return {
            "node_features": [],
            "adjacency_lists": [[] for _ in range(self.num_edge_types)],
            "node_to_graph_map": [],
            "num_graphs_in_batch": 0,
            "num_nodes_in_batch": 0,
        }

    def _add_graph_to_batch(self, raw_batch: Dict[str, Any], graph_sample: GraphSampleType) -> None:
        """Add a graph sample to a minibatch under preparation.

        Args:
            raw_batch: Holder for the currently constructed minibatch (created by _new_batch)
            graph_sample: Graph sample to add.
        """
        num_nodes_in_graph = len(graph_sample.node_features)
        raw_batch["node_features"].extend(graph_sample.node_features)
        raw_batch["node_to_graph_map"].append(
            np.full(
                shape=[num_nodes_in_graph],
                fill_value=raw_batch["num_graphs_in_batch"],
                dtype=np.int32,
            )
        )
        for edge_type_idx, batch_adjacency_list in enumerate(raw_batch["adjacency_lists"]):
            batch_adjacency_list.append(
                graph_sample.adjacency_lists[edge_type_idx].reshape(-1, 2)
                + raw_batch["num_nodes_in_batch"]
            )

    def _finalise_batch(self, raw_batch: Dict[str, Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Turns a raw batch into a minibatch ready to be fed to the model (i.e., converts
        lists to numpy arrays, and concatenate appropriately.

        Args:
            raw_batch: Holder for the currently constructed minibatch (created by _new_batch,
                filled by repeated _add_graph_to_batch calls)

        Returns:
            Pair of batch_features and batch_labels, for use in model.
        """
        batch_features: Dict[str, Any] = {}
        batch_labels: Dict[str, Any] = {}
        batch_features["node_features"] = np.array(raw_batch["node_features"])
        batch_features["node_to_graph_map"] = np.concatenate(raw_batch["node_to_graph_map"])
        batch_features["num_graphs_in_batch"] = raw_batch["num_graphs_in_batch"]
        for i, adjacency_list in enumerate(raw_batch["adjacency_lists"]):
            if len(adjacency_list) > 0:
                batch_features[f"adjacency_list_{i}"] = np.concatenate(adjacency_list)
            else:
                batch_features[f"adjacency_list_{i}"] = np.zeros(shape=(0, 2), dtype=np.int32)

        return batch_features, batch_labels

    def get_batch_tf_data_description(self) -> GraphBatchTFDataDescription:
        """Provides information about dataset-specific data shapes and types
        in the generated minibatch (i.e., additional keys in the features or
        label dictionaries).
        """
        batch_features_types = {
            "node_features": tf.float32,
            "node_to_graph_map": tf.int32,
            "num_graphs_in_batch": tf.int32,
        }
        batch_features_shapes = {
            "node_features": (None,) + self.node_feature_shape,
            "node_to_graph_map": (None,),
            "num_graphs_in_batch": (),
        }
        for edge_type_idx in range(self.num_edge_types):
            batch_features_types[f"adjacency_list_{edge_type_idx}"] = tf.int32
            batch_features_shapes[f"adjacency_list_{edge_type_idx}"] = (None, 2)
        batch_labels_types: Dict[str, Any] = {}
        batch_labels_shapes: Dict[str, Any] = {}

        return GraphBatchTFDataDescription(
            batch_features_types=batch_features_types,
            batch_features_shapes=batch_features_shapes,
            batch_labels_types=batch_labels_types,
            batch_labels_shapes=batch_labels_shapes,
        )

    def get_tensorflow_dataset(
        self, data_fold: DataFold, use_worker_threads: bool = True
    ) -> tf.data.Dataset:
        """Construct a TensorFlow dataset from the _graph_batch_iterator of this class.

        Returns:
            A tensorflow Dataset object. Each element in the dataset is a pair of
            dictionaries representing features and labels.
            The content of these is determined by the _finalise_batch method.
        """
        data_description = self.get_batch_tf_data_description()

        if use_worker_threads:
            graph_batch_iterator = lambda: DoubleBufferedIterator(
                self.graph_batch_iterator(data_fold)
            )
        else:
            graph_batch_iterator = lambda: self.graph_batch_iterator(data_fold)

        dataset = tf.data.Dataset.from_generator(
            generator=graph_batch_iterator,
            output_types=(
                data_description.batch_features_types,
                data_description.batch_labels_types,
            ),
            output_shapes=(
                data_description.batch_features_shapes,
                data_description.batch_labels_shapes,
            ),
        )

        return dataset



"""Reimplementation of the qm9 dataset."""
import inspect
import logging
import os
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Set

import numpy as np
import tensorflow as tf
from dpu_utils.utils import RichPath

from .graph_dataset import DataFold, GraphSample, GraphBatchTFDataDescription, GraphDataset

logger = logging.getLogger(__name__)


class QM9GraphSample(GraphSample):
    """Data structure holding a single QM9 graph."""

    def __init__(
        self,
        adjacency_lists: List[np.ndarray],
        type_to_node_to_num_incoming_edges: np.ndarray,
        node_features: List[np.ndarray],
        target_value: float,
    ):
        super().__init__(adjacency_lists, type_to_node_to_num_incoming_edges, node_features)
        self._target_value = target_value

    @property
    def target_value(self) -> float:
        """Target value of the regression task."""
        return self._target_value

    def __str__(self):
        return (
            f"Adj:            {self._adjacency_lists}\n"
            f"Node_features:  {self._node_features}\n"
            f"Target_values:  {self._target_value}"
        )


class QM9Dataset(GraphDataset[QM9GraphSample]):
    """
    QM9 Dataset class.
    """

    @classmethod
    def get_default_hyperparameters(cls) -> Dict[str, Any]:
        return {
            "max_nodes_per_batch": 10000,
            "add_self_loop_edges": True,
            "tie_fwd_bkwd_edges": True,
            "task_id": 0,
        }

    def __init__(self, params: Dict[str, Any], metadata: Optional[Dict[str, Any]] = None):
        logger.info("Initialising QM9 Dataset.")
        super().__init__(params, metadata=metadata)
        self._params = params
        self._num_fwd_edge_types = 4
        if params["tie_fwd_bkwd_edges"]:
            self._num_edge_types = self._num_fwd_edge_types
        else:
            self._num_edge_types = 2 * self._num_fwd_edge_types
        self._num_edge_types += int(params["add_self_loop_edges"])

        self._node_feature_shape = None
        self._loaded_data: Dict[DataFold, List[QM9GraphSample]] = {}
        logger.debug("Done initialising QM9 dataset.")


    @property
    def num_edge_types(self) -> int:
        return self._num_edge_types

    @classmethod
    def default_data_directory(cls):
        curr_dir = Path(os.path.abspath(inspect.getsourcefile(lambda: 0)))
        data_directory = os.path.join(curr_dir.parent.parent, "data")
        return data_directory

    def load_data(self, path: RichPath, folds_to_load: Optional[Set[DataFold]] = None) -> None:
        """Load the data from disk."""
        if path is None:
            path = RichPath.create(self.default_data_directory())
        logger.info("Starting to load data.")

        # If we haven't defined what folds to load, load all:
        if folds_to_load is None:
            folds_to_load = {DataFold.TRAIN, DataFold.VALIDATION, DataFold.TEST}

        if DataFold.TRAIN in folds_to_load:
            self._loaded_data[DataFold.TRAIN] = self.__load_data(path.join("train.jsonl.gz"))
            logger.debug("Done loading training data.")
        if DataFold.VALIDATION in folds_to_load:
            self._loaded_data[DataFold.VALIDATION] = self.__load_data(path.join("valid.jsonl.gz"))
            logger.debug("Done loading validation data.")
        if DataFold.TEST in folds_to_load:
            self._loaded_data[DataFold.TEST] = self.__load_data(path.join("test.jsonl.gz"))
            logger.debug("Done loading test data.")

    def load_data_from_list(
        self, datapoints: List[Dict[str, Any]], target_fold: DataFold = DataFold.TEST
    ):
        raise NotImplementedError()

    def __load_data(self, data_file: RichPath) -> List[QM9GraphSample]:
        data = list(
            data_file.read_by_file_suffix()
        )  # list() needed for .jsonl case, where .read*() is just a generator
        return self.__process_raw_graphs(data)

    def __process_raw_graphs(self, raw_data: Iterable[Any]) -> List[QM9GraphSample]:
        processed_graphs = []
        for d in raw_data:
            (type_to_adjacency_list, type_to_num_incoming_edges) = self.__graph_to_adjacency_lists(
                d["graph"], num_nodes=len(d["node_features"])
            )
            processed_graphs.append(
                QM9GraphSample(
                    adjacency_lists=type_to_adjacency_list,
                    type_to_node_to_num_incoming_edges=type_to_num_incoming_edges,
                    node_features=d["node_features"],
                    target_value=d["targets"][self.params["task_id"]][0],
                )
            )
        return processed_graphs

    def __graph_to_adjacency_lists(
        self, graph: Iterable[Tuple[int, int, int]], num_nodes: int
    ) -> Tuple[List[np.ndarray], np.ndarray]:
        type_to_adj_list = [
            [] for _ in range(self._num_fwd_edge_types + int(self.params["add_self_loop_edges"]))
        ]  # type: List[List[Tuple[int, int]]]
        type_to_num_incoming_edges = np.zeros(shape=(self.num_edge_types, num_nodes))
        for src, e, dest in graph:
            if self.params["add_self_loop_edges"]:
                fwd_edge_type = e  # 0 will be the self-loop type
            else:
                fwd_edge_type = e - 1  # Make edges start from 0
            type_to_adj_list[fwd_edge_type].append((src, dest))
            type_to_num_incoming_edges[fwd_edge_type, dest] += 1
            if self.params["tie_fwd_bkwd_edges"]:
                type_to_adj_list[fwd_edge_type].append((dest, src))
                type_to_num_incoming_edges[fwd_edge_type, src] += 1

        if self.params["add_self_loop_edges"]:
            # Add self-loop edges (idx 0, which isn't used in the data):
            for node in range(num_nodes):
                type_to_num_incoming_edges[0, node] = 1
                type_to_adj_list[0].append((node, node))

        # Add backward edges as an additional edge type that goes backwards:
        if not (self.params["tie_fwd_bkwd_edges"]):
            # for (edge_type, adj_list) in enumerate(type_to_adj_list):
            num_edge_types_in_adj_lists = len(type_to_adj_list)
            for edge_type in range(num_edge_types_in_adj_lists):
                adj_list = type_to_adj_list[edge_type]
                # Don't add self loops again!
                if edge_type == 0 and self.params["add_self_loop_edges"]:
                    continue
                bkwd_edge_type = len(type_to_adj_list)
                type_to_adj_list.append([(y, x) for (x, y) in adj_list])
                for (x, y) in adj_list:
                    type_to_num_incoming_edges[bkwd_edge_type][y] += 1

        # Convert the adjacency lists to numpy arrays.
        type_to_adj_list = [
            np.array(adj_list, dtype=np.int32)
            if len(adj_list) > 0
            else np.zeros(shape=(0, 2), dtype=np.int32)
            for adj_list in type_to_adj_list
        ]
        return type_to_adj_list, type_to_num_incoming_edges

    @property
    def node_feature_shape(self) -> Tuple:
        """Return the shape of the node features."""
        if self._node_feature_shape is None:
            some_data_fold = next(iter(self._loaded_data.values()))
            self._node_feature_shape = (len(some_data_fold[0].node_features[0]),)
        return self._node_feature_shape

    def _graph_iterator(self, data_fold: DataFold) -> Iterator[QM9GraphSample]:
        logger.debug("QM9 graph iterator requested.")
        loaded_data = self._loaded_data[data_fold]
        if data_fold == DataFold.TRAIN:
            np.random.shuffle(loaded_data)
        return iter(loaded_data)

    def _new_batch(self) -> Dict[str, Any]:
        new_batch = super()._new_batch()
        new_batch["target_value"] = []
        return new_batch

    def _add_graph_to_batch(self, raw_batch: Dict[str, Any], graph_sample: QM9GraphSample) -> None:
        super()._add_graph_to_batch(raw_batch, graph_sample)
        raw_batch["target_value"].append(graph_sample.target_value)

    def _finalise_batch(self, raw_batch) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        batch_features, batch_labels = super()._finalise_batch(raw_batch)
        return batch_features, {"target_value": raw_batch["target_value"]}

    def get_batch_tf_data_description(self) -> GraphBatchTFDataDescription:
        data_description = super().get_batch_tf_data_description()
        return GraphBatchTFDataDescription(
            batch_features_types=data_description.batch_features_types,
            batch_features_shapes=data_description.batch_features_shapes,
            batch_labels_types={**data_description.batch_labels_types, "target_value": tf.float32},
            batch_labels_shapes={**data_description.batch_labels_shapes, "target_value": (None,)},
        )


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



"""Message passing layer."""
from abc import abstractmethod
from typing import Dict, List, NamedTuple, Tuple, Any

import tensorflow as tf

from tf2_gnn.utils.param_helpers import get_activation_function, get_aggregation_function


class MessagePassingInput(NamedTuple):
    """A named tuple to hold input to the message passing layer."""

    node_embeddings: tf.Tensor
    adjacency_lists: Tuple[tf.Tensor, ...]


class MessagePassing(tf.keras.layers.Layer):
    """Abstract class to compute new graph states by neural message passing.

    Users should create a specific type of message passing layer by:
        * Implementing the abstract method `_message_function`, which performs the calculation of
          the messages that should be sent around the graph.
        * (Optional) Overriding the `_aggregation_function` method, which calculate the new (pre-
          activation) node state by aggregating the messages that were sent to the node. The built
          in defaults for this method are sum, mean, max and sqrt_n, which can be chosen by setting
          the "aggregation_function" setting the parameter dictionary.


    Throughout we use the following abbreviations in shape descriptions:
        * V: number of nodes
        * D: state dimension
        * L: number of different edge types
        * E: number of edges of a given edge type
        * D: input node representation dimension
        * H: output node representation dimension (set as hidden_dim)
    """

    @classmethod
    def get_default_hyperparameters(cls):
        return {
            "aggregation_function": "sum",  # One of sum, mean, max, sqrt_n
            "message_activation_function": "relu",  # One of relu, leaky_relu, elu, gelu, tanh
            "hidden_dim": 7,
        }

    def __init__(self, params: Dict[str, Any], **kwargs):
        super().__init__(**kwargs)
        self._hidden_dim = int(params["hidden_dim"])

        aggregation_fn_name = params["aggregation_function"]
        self._aggregation_fn = get_aggregation_function(aggregation_fn_name)

        activation_fn_name = params["message_activation_function"]
        self._activation_fn = get_activation_function(activation_fn_name)

    @abstractmethod
    def _message_function(
        self,
        edge_source_states: tf.Tensor,
        edge_target_states: tf.Tensor,
        num_incoming_to_node_per_message: tf.Tensor,
        edge_type_idx: int,
        training: bool,
    ) -> tf.Tensor:
        """Abstract method to calculate the messages passed from a source nodes to target nodes.

        NOTE: This function should calculate messages for a single edge type.

        Args:
            edge_source_states: A tensor of shape [E, D] giving the node states at the source node
                of each edge.
            edge_target_states: A tensor of shape [E, D] giving the node states at the target node
                of each edge.
            num_incoming_to_node_per_message: A tensor of shape [E] giving the number of messages
                entering the target node of each edge. For example, if
                num_incoming_to_node_per_message[i] = 4, then there are 4 messages whose target
                node is the target node of message i.
            edge_type_idx: The edge type that that these messages correspond to.

        Returns:
            If this layer is to be used with the `_compute_new_node_embeddings` function as defined
            in this class, this function should return a tensor of shape [E, H] representing the
            messages passed along each edge.
        """
        pass

    def call(self, inputs: MessagePassingInput, training: bool = False):
        """Call the message passing layer.

        Args:
            inputs: A tuple containing two items:
                node_embeddings: float32 tensor of shape [V, D], the original representation of each
                    node in the graph.
                adjacency_lists: Tuple of L adjacency lists, represented as int32 tensors of shape
                    [E, 2]. Concretely, adjacency_lists[l][k,:] == [v, u] means that the k-th edge
                    of type l connects node v to node u.
            training: A bool that denotes whether we are in training mode.

        Returns:
            float32 tensor of shape [V, hidden_dim]
        """
        node_embeddings, adjacency_lists = inputs.node_embeddings, inputs.adjacency_lists
        num_nodes = tf.shape(node_embeddings)[0]

        messages_per_type = self._calculate_messages_per_type(
            adjacency_lists, node_embeddings, training
        )

        edge_type_to_message_targets = [
            adjacency_list_for_edge_type[:, 1] for adjacency_list_for_edge_type in adjacency_lists
        ]

        new_node_states = self._compute_new_node_embeddings(
            node_embeddings, messages_per_type, edge_type_to_message_targets, num_nodes, training,
        )  # Shape [V, H]

        return new_node_states

    def _compute_new_node_embeddings(
        self,
        cur_node_embeddings: tf.Tensor,
        messages_per_type: List[tf.Tensor],
        edge_type_to_message_targets: List[tf.Tensor],
        num_nodes: tf.Tensor,
        training: bool,
    ):
        """Aggregate the messages using the aggregation function specified in the params dict.

        If more exotic types of aggregation are required (such as attention methods, etc.), this
        method should be overwritten in the child class.

        NOTE: If you are overriding this method definition, the `messages` input is a list
        containing the return value of the `_message_function` method for each edge type.

        Args:
            cur_node_embeddings: Current node embeddings as a float32 tensor of shape [V, H]
            messages_per_type: A list of messages to be aggregated with length L. Element i of the
            list should be a tensor of the messages for edge type i, with shape [E, H]
            edge_type_to_message_targets: A list of tensors containing the target nodes of each
                message, for each edge type. Each tensor in the list should have shape [E]. For
                example, edge_type_to_message_targets[i][j] = v means that the jth message of edge
                type i is being sent to node v.
            num_nodes: The total number of nodes in the graph, V.

        Returns:
            A tensor of shape [V, H] representing the aggregated messages for each node.

        """
        # Let M be the number of messages (sum of all E):
        message_targets = tf.concat(edge_type_to_message_targets, axis=0)  # Shape [M]
        messages = tf.concat(messages_per_type, axis=0)  # Shape [M, H]

        aggregated_messages = self._aggregation_fn(
            data=messages, segment_ids=message_targets, num_segments=num_nodes
        )
        return self._activation_fn(aggregated_messages)

    def _calculate_messages_per_type(
        self,
        adjacency_lists: Tuple[tf.Tensor, ...],
        node_embeddings: tf.Tensor,
        training: bool = False,
    ) -> List[tf.Tensor]:
        messages_per_type = []  # list of tensors of messages of shape [E, H]

        # Calculate this once.
        type_to_num_incoming_edges = calculate_type_to_num_incoming_edges(
            node_embeddings, adjacency_lists
        )
        # Collect incoming messages per edge type
        for edge_type_idx, adjacency_list_for_edge_type in enumerate(adjacency_lists):
            edge_sources = adjacency_list_for_edge_type[:, 0]
            edge_targets = adjacency_list_for_edge_type[:, 1]
            edge_source_states = tf.nn.embedding_lookup(
                params=node_embeddings, ids=edge_sources
            )  # Shape [E, H]
            edge_target_states = tf.nn.embedding_lookup(
                params=node_embeddings, ids=edge_targets
            )  # Shape [E, H]

            num_incoming_to_node_per_message = tf.nn.embedding_lookup(
                params=type_to_num_incoming_edges[edge_type_idx, :], ids=edge_targets
            )  # Shape [E, H]

            # Calculate the messages themselves:
            messages = self._message_function(
                edge_source_states,
                edge_target_states,
                num_incoming_to_node_per_message,
                edge_type_idx,
                training,
            )

            messages_per_type.append(messages)
        return messages_per_type


MESSAGE_PASSING_IMPLEMENTATIONS: Dict[str, MessagePassing] = {}

def register_message_passing_implementation(cls):
    """Decorator used to register a message passing class implementation"""
    MESSAGE_PASSING_IMPLEMENTATIONS[cls.__name__.lower()] = cls
    return cls


def calculate_type_to_num_incoming_edges(node_embeddings, adjacency_lists):
    """Calculate the type_to_num_incoming_edges tensor.

        Returns:
            float32 tensor of shape [L, V] representing the number of incoming edges of
            a given type. Concretely, type_to_num_incoming_edges[l, v] is the number of
            edge of type l connecting to node v.

    >>> node_embeddings = tf.random.normal(shape=(5, 3))
    >>> adjacency_lists = [
    ...    tf.constant([[0, 1], [2, 4], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[2, 3], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[3, 1]], dtype=tf.int32),
    ... ]
    ...
    >>> print(calculate_type_to_num_incoming_edges(node_embeddings, adjacency_lists))
    tf.Tensor(
    [[0. 1. 0. 0. 2.]
     [0. 0. 0. 1. 1.]
     [0. 1. 0. 0. 0.]], shape=(3, 5), dtype=float32)
    """

    type_to_num_incoming_edges = []
    for edge_type_adjacency_list in adjacency_lists:
        targets = edge_type_adjacency_list[:, 1]
        indices = tf.expand_dims(targets, axis=-1)
        num_incoming_edges = tf.scatter_nd(
            indices=indices,
            updates=tf.ones_like(targets, dtype=tf.float32),
            shape=(tf.shape(node_embeddings)[0],),
        )
        type_to_num_incoming_edges.append(num_incoming_edges)

    return tf.stack(type_to_num_incoming_edges)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""Relation Graph Isomorphism Network message propogation layer."""
from typing import Dict, List, Any, Optional

import tensorflow as tf
from dpu_utils.tf2utils import MLP

from .message_passing import MessagePassing, MessagePassingInput, register_message_passing_implementation
from .gnn_edge_mlp import GNN_Edge_MLP
from tf2_gnn.utils.constants import SMALL_NUMBER


@register_message_passing_implementation
class RGIN(GNN_Edge_MLP):
    """Compute new graph states by neural message passing using MLPs for state updates
    and message computation.
    For this, we assume existing node states h^t_v and a list of per-edge-type adjacency
    matrices A_\ell.

    We compute new states as follows:
        h^{t+1}_v := \sigma(MLP_{aggr}(\sum_\ell \sum_{(u, v) \in A_\ell} MLP_\ell(h^t_u)))
    The learnable parameters of this are the MLPs MLP_\ell.
    This is derived from Cor. 6 of arXiv:1810.00826, instantiating the functions f, \phi
    with _separate_ MLPs. This is more powerful than the GIN formulation in Eq. (4.1) of
    arXiv:1810.00826, as we want to be able to distinguish graphs of the form
     G_1 = (V={1, 2, 3}, E_1={(1, 2)}, E_2={(3, 2)})
    and
     G_2 = (V={1, 2, 3}, E_1={(3, 2)}, E_2={(1, 2)})
    from each other. If we would treat all edges the same,
    G_1.E_1 \cup G_1.E_2 == G_2.E_1 \cup G_2.E_2 would imply that the two graphs
    become indistuingishable.
    Hence, we introduce per-edge-type MLPs, which also means that we have to drop
    the optimisation of modelling f \circ \phi by a single MLP used in the original
    GIN formulation.

    Note that RGIN is implemented as a special-case of GNN_Edge_MLP, setting some
    different default hyperparameters and adding a different message aggregation
    function, but re-using the message passing functionality.

    We use the following abbreviations in shape descriptions:
    * V: number of nodes
    * L: number of different edge types
    * E: number of edges of a given edge type
    * D: input node representation dimension
    * H: output node representation dimension (set as hidden_dim)

    >>> node_embeddings = tf.random.normal(shape=(5, 3))
    >>> adjacency_lists = (
    ...    tf.constant([[0, 1], [2, 4], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[2, 3], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[3, 1]], dtype=tf.int32),
    ... )
    ...
    >>> params = RGIN.get_default_hyperparameters()
    >>> params["hidden_dim"] = 12
    >>> layer = RGIN(params)
    >>> output = layer(MessagePassingInput(node_embeddings, adjacency_lists))
    >>> print(output)
    tf.Tensor(..., shape=(5, 12), dtype=float32)
    """

    @classmethod
    def get_default_hyperparameters(cls):
        these_hypers = {
            "use_target_state_as_input": False,
            "num_edge_MLP_hidden_layers": 1,
            "num_aggr_MLP_hidden_layers": None,
        }
        gnn_edge_mlp_hypers = super().get_default_hyperparameters()
        gnn_edge_mlp_hypers.update(these_hypers)
        return gnn_edge_mlp_hypers

    def __init__(self, params: Dict[str, Any], **kwargs):
        super().__init__(params, **kwargs)
        self._num_aggr_MLP_hidden_layers: Optional[int] = params["num_aggr_MLP_hidden_layers"]
        self._aggregation_mlp: Optional[MLP] = None

    def build(self, input_shapes: MessagePassingInput):
        node_embedding_shapes = input_shapes.node_embeddings
        if self._num_aggr_MLP_hidden_layers is not None:
            with tf.name_scope("aggregation_MLP"):
                self._aggregation_mlp = MLP(
                    out_size=self._hidden_dim,
                    hidden_layers=[self._hidden_dim] * self._num_aggr_MLP_hidden_layers,
                )
                self._aggregation_mlp.build(tf.TensorShape((None, self._hidden_dim)))
        super().build(input_shapes)

    def _compute_new_node_embeddings(
        self,
        cur_node_embeddings: tf.Tensor,
        messages_per_type: List[tf.Tensor],
        edge_type_to_message_targets: List[tf.Tensor],
        num_nodes: tf.Tensor,
        training: bool,
    ):
        # Let M be the number of messages (sum of all E):
        message_targets = tf.concat(edge_type_to_message_targets, axis=0)  # Shape [M]
        messages = tf.concat(messages_per_type, axis=0)  # Shape [M, H]

        aggregated_messages = self._aggregation_fn(
            data=messages, segment_ids=message_targets, num_segments=num_nodes
        )
        if self._aggregation_mlp is not None:
            aggregated_messages = self._aggregation_mlp(aggregated_messages, training)

        return self._activation_fn(aggregated_messages)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from .nodes_to_graph_representation import (
    WeightedSumGraphRepresentation,
    NodesToGraphRepresentationInput,
)
from .graph_global_exchange import (
    GraphGlobalExchangeInput,
    GraphGlobalExchange,
    GraphGlobalMeanExchange,
    GraphGlobalGRUExchange,
    GraphGlobalMLPExchange,
)
from .gnn import GNNInput, GNN
from .message_passing import get_known_message_passing_classes



import time
from abc import abstractmethod
from typing import Tuple, List, Dict, Optional, Any, Iterable, Union

import tensorflow as tf

from tf2_gnn import GNNInput, GNN
from tf2_gnn.data import GraphDataset


class GraphTaskModel(tf.keras.Model):
    @classmethod
    def get_default_hyperparameters(cls, mp_style: Optional[str] = None) -> Dict[str, Any]:
        """Get the default hyperparameter dictionary for the class."""
        params = {f"gnn_{name}": value for name, value in GNN.get_default_hyperparameters(mp_style).items()}
        these_hypers: Dict[str, Any] = {
            "optimizer": "Adam",  # One of "SGD", "RMSProp", "Adam"
            "learning_rate": 0.001,
            "learning_rate_decay": 0.98,
            "momentum": 0.85,
            "gradient_clip_value": 1.0,
            "use_intermediate_gnn_results": False,
        }
        params.update(these_hypers)
        return params

    def __init__(self, params: Dict[str, Any], dataset: GraphDataset, name: str = None):
        super().__init__(name=name)
        self._params = params
        self._num_edge_types = dataset.num_edge_types
        self._use_intermediate_gnn_results = params.get("use_intermediate_gnn_results", False)
        self._train_step_counter = 0

    def build(self, input_shapes: Dict[str, Any]):
        graph_params = {
            name[4:]: value for name, value in self._params.items() if name.startswith("gnn_")
        }
        self._gnn = GNN(graph_params)
        self._gnn.build(
            GNNInput(
                node_features=self.get_initial_node_feature_shape(input_shapes),
                adjacency_lists=tuple(
                    input_shapes[f"adjacency_list_{edge_type_idx}"]
                    for edge_type_idx in range(self._num_edge_types)
                ),
                node_to_graph_map=tf.TensorShape((None,)),
                num_graphs=tf.TensorShape(()),
            )
        )

        super().build([])

    def get_initial_node_feature_shape(self, input_shapes) -> tf.TensorShape:
        return input_shapes["node_features"]

    def compute_initial_node_features(self, inputs, training: bool) -> tf.Tensor:
        return inputs["node_features"]

    @abstractmethod
    def compute_task_output(
        self,
        batch_features: Dict[str, tf.Tensor],
        final_node_representations: Union[tf.Tensor, Tuple[tf.Tensor, Tuple[tf.Tensor, ...]]],
        training: bool,
    ) -> Any:
        """Compute task-specific output (labels, scores, regression values, ...).

        Args:
            batch_features: Input data for minibatch (as generated by the used datasets
                _finalise_batch method).
            final_node_representations:
                Per default (or if the hyperparameter "use_intermediate_gnn_results" was
                set to False), the final representations of the graph nodes as computed
                by the GNN.
                If the hyperparameter "use_intermediate_gnn_results" was set to True,
                a pair of the final node representation and all intermediate node
                representations, including the initial one.
            training: Flag indicating if we are training or not.

        Returns:
            Implementor's choice, but will be passed as task_output to compute_task_metrics
            during training/evaluation.
        """
        pass

    def compute_final_node_representations(self, inputs, training: bool):
        # Pack input data from keys back into a tuple:
        adjacency_lists: Tuple[tf.Tensor, ...] = tuple(
            inputs[f"adjacency_list_{edge_type_idx}"]
            for edge_type_idx in range(self._num_edge_types)
        )

        # Start the model computations:
        initial_node_features = self.compute_initial_node_features(inputs, training)
        gnn_input = GNNInput(
            node_features=initial_node_features,
            adjacency_lists=adjacency_lists,
            node_to_graph_map=inputs["node_to_graph_map"],
            num_graphs=inputs["num_graphs_in_batch"],
        )

        gnn_output = self._gnn(
            gnn_input,
            training=training,
            return_all_representations=self._use_intermediate_gnn_results
        )
        return gnn_output

    def call(self, inputs, training: bool):
        final_node_representations = self.compute_final_node_representations(inputs, training)
        return self.compute_task_output(inputs, final_node_representations, training)

    @abstractmethod
    def compute_task_metrics(
        self,
        batch_features: Dict[str, tf.Tensor],
        task_output: Any,
        batch_labels: Dict[str, tf.Tensor],
    ) -> Dict[str, tf.Tensor]:
        """Compute task-specific loss & metrics (accuracy, F1 score, ...)

        Args:
            batch_features: Input data for minibatch (as generated by the used datasets
                _finalise_batch method).
            task_output: Output generated by compute_task_output.
            batch_labels: Target labels for minibatch (as generated by the used datasets
                _finalise_batch method).

        Returns:
            Dictionary of different metrics. Has to contain value for key
            "loss" (which will be used during training as starting point for backprop).
        """
        pass

    @abstractmethod
    def compute_epoch_metrics(self, task_results: List[Any]) -> Tuple[float, str]:
        """Compute single value used to measure quality of model at one epoch, where
        lower is better.
        This value, computed on the validation set, is used to determine if model
        training is still improving results.

        Args:
            task_results: List of results obtained by compute_task_metrics for the
                batches in one epoch.

        Returns:
            Pair of a metric value (lower ~ better) and a human-readable string
            describing it.
        """
        pass

    def _make_optimizer(
        self,
        learning_rate: Optional[
            Union[float, tf.keras.optimizers.schedules.LearningRateSchedule]
        ] = None,
    ) -> tf.keras.optimizers.Optimizer:
        """
        Create fresh optimizer.

        Args:
            learning_rate: Optional setting for learning rate; if unset, will
                use value from self._params["learning_rate"].
        """
        if learning_rate is None:
            learning_rate = self._params["learning_rate"]

        optimizer_name = self._params["optimizer"].lower()
        if optimizer_name == "sgd":
            return tf.keras.optimizers.SGD(
                learning_rate=learning_rate,
                momentum=self._params["momentum"],
                clipvalue=self._params["gradient_clip_value"],
            )
        elif optimizer_name == "rmsprop":
            return tf.keras.optimizers.RMSprop(
                learning_rate=learning_rate,
                decay=self._params["learning_rate_decay"],
                momentum=self._params["momentum"],
                clipvalue=self._params["gradient_clip_value"],
            )
        elif optimizer_name == "adam":
            return tf.keras.optimizers.Adam(
                learning_rate=learning_rate,
                clipvalue=self._params["gradient_clip_value"],
            )
        else:
            raise Exception('Unknown optimizer "%s".' % (self._params["optimizer"]))

    def _apply_gradients(
        self, gradient_variable_pairs: Iterable[Tuple[tf.Tensor, tf.Variable]]
    ) -> None:
        """
        Apply gradients to the models variables during training.

        Args:
            gradient_variable_pairs: Iterable of pairs of gradients for a variable
                and the variable itself. Suitable to be fed into
                tf.keras.optimizer.*.apply_gradients.
        """
        if getattr(self, "_optimizer", None) is None:
            self._optimizer = self._make_optimizer()

        self._optimizer.apply_gradients(gradient_variable_pairs)

    # ----------------------------- Training Loop
    def run_one_epoch(
        self, dataset: tf.data.Dataset, quiet: bool = False, training: bool = True,
    ) -> Tuple[float, float, List[Any]]:
        epoch_time_start = time.time()
        total_num_graphs = 0
        task_results = []
        total_loss = tf.constant(0, dtype=tf.float32)
        for step, (batch_features, batch_labels) in enumerate(dataset):
            with tf.GradientTape() as tape:
                task_output = self(batch_features, training=training)
                task_metrics = self.compute_task_metrics(batch_features, task_output, batch_labels)
            total_loss += task_metrics["loss"]
            total_num_graphs += batch_features["num_graphs_in_batch"]
            task_results.append(task_metrics)

            if training:
                gradients = tape.gradient(
                    task_metrics["loss"], self.trainable_variables
                )
                self._apply_gradients(zip(gradients, self.trainable_variables))
                self._train_step_counter += 1

            if not quiet:
                epoch_graph_average_loss = (total_loss / float(total_num_graphs)).numpy()
                batch_graph_average_loss = task_metrics["loss"] / float(batch_features["num_graphs_in_batch"])
                steps_per_second = step / (time.time() - epoch_time_start)
                print(
                    f"   Step: {step:4d}"
                    f"  |  Epoch graph avg. loss = {epoch_graph_average_loss:.5f}"
                    f"  |  Batch graph avg. loss = {batch_graph_average_loss:.5f}"
                    f"  |  Steps per sec = {steps_per_second:.5f}",
                    end="\r"
                )
        if not quiet:
            print("\r\x1b[K", end="")
        total_time = time.time() - epoch_time_start
        return total_loss / float(total_num_graphs), float(total_num_graphs) / total_time, task_results

    # ----------------------------- Prediction Loop
    def predict(self, dataset: tf.data.Dataset):
        task_outputs = []
        for batch_features, _ in dataset:
            task_outputs.append(self(batch_features, training=False))

        # Note: This assumes that the task output is a tensor (true for classification, regression,
        #  etc.) but subclasses implementing more complicated outputs will need to override this.
        return tf.concat(task_outputs, axis=0)



from typing import Dict, Any, NamedTuple, Type, Iterable, Tuple

from ..models import (
    GraphTaskModel,
    NodeMulticlassTask,
    GraphRegressionTask,
    GraphBinaryClassificationTask,
)
from ..data import GraphDataset, JsonLGraphPropertyDataset, QM9Dataset, PPIDataset


class TaskInfo(NamedTuple):
    """A named tuple to hold information about a task."""

    name: str
    dataset_class: Type[GraphDataset]
    dataset_default_hypers: Dict[str, Any]
    model_class: Type[GraphTaskModel]
    model_default_hypers: Dict[str, Any]


TASK_NAME_TO_DATASET_AND_MODEL_INFO: Dict[str, TaskInfo] = {}


def register_task(
    task_name, dataset_class, dataset_default_hypers, model_class, model_default_hypers
):
    TASK_NAME_TO_DATASET_AND_MODEL_INFO[task_name.lower()] = TaskInfo(
        name=task_name,
        dataset_class=dataset_class,
        dataset_default_hypers=dataset_default_hypers,
        model_class=model_class,
        model_default_hypers=model_default_hypers,
    )


def clear_known_tasks() -> None:
    TASK_NAME_TO_DATASET_AND_MODEL_INFO.clear()


def get_known_tasks() -> Iterable[str]:
    for task_info in TASK_NAME_TO_DATASET_AND_MODEL_INFO.values():
        yield task_info.name


def task_name_to_dataset_class(name: str) -> Tuple[Type[GraphDataset], Dict[str, Any]]:
    """
    Map task name to a dataset class and default hyperparameters for that class.
    """
    task_info = TASK_NAME_TO_DATASET_AND_MODEL_INFO.get(name.lower())
    if task_info is None:
        raise ValueError("Unknown task type '%s'" % name)
    return task_info.dataset_class, task_info.dataset_default_hypers


def task_name_to_model_class(name: str) -> Tuple[Type[GraphTaskModel], Dict[str, Any]]:
    """
    Map task name to a model class and default hyperparameters for that class.
    """
    task_info = TASK_NAME_TO_DATASET_AND_MODEL_INFO.get(name.lower())
    if task_info is None:
        raise ValueError("Unknown task type '%s'" % name)
    return task_info.model_class, task_info.model_default_hypers


# Register some default tasks:
register_task(
    task_name="PPI",
    dataset_class=PPIDataset,
    dataset_default_hypers={},
    model_class=NodeMulticlassTask,
    model_default_hypers={},
)

register_task(
    task_name="QM9",
    dataset_class=QM9Dataset,
    dataset_default_hypers={},
    model_class=GraphRegressionTask,
    model_default_hypers={},
)

register_task(
    task_name="GraphRegression",
    dataset_class=JsonLGraphPropertyDataset,
    dataset_default_hypers={"threshold_for_classification": None},
    model_class=GraphRegressionTask,
    model_default_hypers={},
)

register_task(
    task_name="GraphBinaryClassification",
    dataset_class=JsonLGraphPropertyDataset,
    dataset_default_hypers={"threshold_for_classification": 23.0},
    model_class=GraphBinaryClassificationTask,
    model_default_hypers={},
)



import json
import os
import pickle
from typing import Dict, Any, Optional, Set, Type

import h5py
import numpy as np
import tensorflow as tf
from dpu_utils.utils import RichPath
from tensorflow.python.keras import backend as K

from ..data import DataFold, GraphDataset
from ..models import GraphTaskModel
from .dataset_utils import get_dataset, get_model_file_path
from .task_utils import task_name_to_model_class
from .param_helpers import override_model_params_with_hyperdrive_params


def save_model(save_file: str, model: GraphTaskModel, dataset: GraphDataset) -> None:
    data_to_store = {
        "model_class": model.__class__,
        "model_params": model._params,
        "dataset_class": dataset.__class__,
        "dataset_params": dataset._params,
        "dataset_metadata": dataset._metadata,
        "num_edge_types": dataset.num_edge_types,
        "node_feature_shape": dataset.node_feature_shape,
    }
    pkl_file = get_model_file_path(save_file, "pkl")
    hdf5_file = get_model_file_path(save_file, "hdf5")
    with open(pkl_file, "wb") as out_file:
        pickle.dump(data_to_store, out_file, pickle.HIGHEST_PROTOCOL)
    model.save_weights(hdf5_file, save_format="h5")
    print(f"   (Stored model metadata to {pkl_file} and weights to {hdf5_file})")


def load_weights_verbosely(
    save_file: str,
    model: GraphTaskModel,
    warn_about_initialisations: bool = True,
    warn_about_ignored: bool = True,
):
    hdf5_save_file = get_model_file_path(save_file, "hdf5")
    var_name_to_variable = {}
    var_names_unique = True
    for var in model.variables:
        if var.name in var_name_to_variable:
            print(
                f"E: More than one variable with name {var.name} used in model. Please use appropriate name_scopes!"
            )
            var_names_unique = False
        else:
            var_name_to_variable[var.name] = var
    if not var_names_unique:
        raise ValueError(
            "Model variables have duplicate names, making weight restoring impossible."
        )

    var_name_to_weights = {}

    def hdf5_item_visitor(name, item):
        if not isinstance(item, h5py.Dataset):
            return
        if name in var_name_to_weights:
            raise ValueError(
                f"More than one variable with name {name} used in hdf5 file. Please use appropriate name_scopes!"
            )
        else:
            var_name_to_weights[name] = np.array(item)

    with h5py.File(hdf5_save_file, mode="r") as data_hdf5:
        # For some reason, the first layer of attributes is auto-generated names instead of actual names:
        for model_sublayer in data_hdf5.values():
            model_sublayer.visititems(hdf5_item_visitor)

    tfvar_weight_tuples = []
    for var_name, tfvar in var_name_to_variable.items():
        saved_weight = var_name_to_weights.get(var_name)
        if saved_weight is None:
            if warn_about_initialisations:
                print(f"I: Weights for {var_name} freshly initialised.")
        else:
            tfvar_weight_tuples.append((tfvar, saved_weight))

    if warn_about_ignored:
        for var_name in var_name_to_weights.keys():
            if var_name not in var_name_to_variable:
                print(f"I: Model does not use saved weights for {var_name}.")

    K.batch_set_value(tfvar_weight_tuples)


def load_dataset_for_prediction(trained_model_file: str):
    with open(get_model_file_path(trained_model_file, "pkl"), "rb") as in_file:
        data_to_load = pickle.load(in_file)
    dataset_class: Type[GraphDataset] = data_to_load["dataset_class"]

    return dataset_class(
        params=data_to_load.get("dataset_params", {}),
        metadata=data_to_load.get("dataset_metadata", {}),
    )


def load_model_for_prediction(trained_model_file: str, dataset: GraphDataset):
    with open(get_model_file_path(trained_model_file, "pkl"), "rb") as in_file:
        data_to_load = pickle.load(in_file)
    model_class: Type[GraphTaskModel] = data_to_load["model_class"]

    # Clear the Keras session so that unique naming does not mess up weight loading.
    tf.keras.backend.clear_session()

    model = model_class(params=data_to_load.get("model_params", {}), dataset=dataset,)

    data_description = dataset.get_batch_tf_data_description()
    model.build(data_description.batch_features_shapes)

    print(f"Restoring model weights from {trained_model_file}.")
    load_weights_verbosely(trained_model_file, model)

    return model


def get_model(
    msg_passing_implementation: str,
    task_name: str,
    model_cls: Type[GraphTaskModel],
    dataset: GraphDataset,
    dataset_model_optimised_default_hyperparameters: Dict[str, Any],
    loaded_model_hyperparameters: Dict[str, Any],
    cli_model_hyperparameter_overrides: Dict[str, Any],
    hyperdrive_hyperparameter_overrides: Dict[str, str],
) -> GraphTaskModel:
    if not model_cls:
        model_cls, model_default_hyperparameter_overrides = task_name_to_model_class(
            task_name
        )
        model_params = model_cls.get_default_hyperparameters(msg_passing_implementation)
        print(f" Model default parameters: {model_params}")
        model_params.update(model_default_hyperparameter_overrides)
        if len(model_default_hyperparameter_overrides):
            print(
                f"  Model parameters overridden by task defaults: {model_default_hyperparameter_overrides}"
            )
        model_params.update(dataset_model_optimised_default_hyperparameters)
        if len(dataset_model_optimised_default_hyperparameters):
            print(
                f"  Model parameters overridden by task/model defaults: {dataset_model_optimised_default_hyperparameters}"
            )
    else:
        model_params = loaded_model_hyperparameters
    model_params.update(cli_model_hyperparameter_overrides)
    if len(cli_model_hyperparameter_overrides):
        print(
            f"  Model parameters overridden from CLI: {cli_model_hyperparameter_overrides}"
        )
    if len(hyperdrive_hyperparameter_overrides) > 0:
        override_model_params_with_hyperdrive_params(
            model_params, hyperdrive_hyperparameter_overrides
        )
        print(
            f"  Model parameters overridden for Hyperdrive: {hyperdrive_hyperparameter_overrides}"
        )
    return model_cls(model_params, dataset=dataset)


# TODO: A better solution to 'loading weights only without model and class' is required.
# In particular, need to ensure that the weights and the proposed model to be trained match up in their
# base components (i.e. if loading GNN weights, dimensions need to match for the GNN in finetuning)
def get_model_and_dataset(
    task_name: Optional[str],
    msg_passing_implementation: Optional[str],
    data_path: RichPath,
    trained_model_file: Optional[str],
    cli_data_hyperparameter_overrides: Optional[str],
    cli_model_hyperparameter_overrides: Optional[str],
    hyperdrive_hyperparameter_overrides: Dict[str, str] = {},
    folds_to_load: Optional[Set[DataFold]] = None,
    load_weights_only: bool = False,
):
    # case of a trained model file being passed, where the entire model should be loaded,
    # a new class and dataset type are not required
    if trained_model_file and not load_weights_only:
        with open(get_model_file_path(trained_model_file, "pkl"), "rb") as in_file:
            data_to_load = pickle.load(in_file)
        model_class = data_to_load["model_class"]
        dataset_class = data_to_load["dataset_class"]
        default_task_model_hypers = {}
    # case 1: trained_model_file and loading weights only -- create new dataset and class of the type specified by the
    # task to be trained, but use weights from another corresponding model
    # case 2: no model to be loaded; make fresh dataset and model classes
    elif (trained_model_file and load_weights_only) or not trained_model_file:
        data_to_load = {}
        model_class, dataset_class = None, None

        # Load potential task-specific defaults:
        default_task_model_hypers = {}
        task_model_default_hypers_file = os.path.join(
            os.path.dirname(__file__),
            "default_hypers",
            "%s_%s.json" % (task_name, msg_passing_implementation),
        )
        print(
            f"Trying to load task/model-specific default parameters from {task_model_default_hypers_file} ... ",
            end="",
        )
        if os.path.exists(task_model_default_hypers_file):
            print("File found.")
            with open(task_model_default_hypers_file, "rt") as f:
                default_task_model_hypers = json.load(f)
        else:
            print("File not found, using global defaults.")

        if not trained_model_file and load_weights_only:
            raise ValueError(
                "Cannot load only weights when model file from which to load is not specified."
            )

    dataset = get_dataset(
        task_name,
        dataset_class,
        default_task_model_hypers.get("task_params", {}),
        data_to_load.get("dataset_params", {}),
        json.loads(cli_data_hyperparameter_overrides or "{}"),
        data_to_load.get("dataset_metadata", {}),
    )

    # Actually load data:
    print(f"Loading data from {data_path}.")
    dataset.load_data(data_path, folds_to_load)

    model = get_model(
        msg_passing_implementation,
        task_name,
        model_class,
        dataset,
        dataset_model_optimised_default_hyperparameters=default_task_model_hypers.get(
            "model_params", {}
        ),
        loaded_model_hyperparameters=data_to_load.get("model_params", {}),
        cli_model_hyperparameter_overrides=json.loads(
            cli_model_hyperparameter_overrides or "{}"
        ),
        hyperdrive_hyperparameter_overrides=hyperdrive_hyperparameter_overrides or {},
    )

    data_description = dataset.get_batch_tf_data_description()
    model.build(data_description.batch_features_shapes)

    # If needed, load weights for model:
    if trained_model_file:
        print(f"Restoring model weights from {trained_model_file}.")
        load_weights_verbosely(trained_model_file, model)

    return dataset, model



from abc import abstractmethod
from typing import NamedTuple

import tensorflow as tf
from dpu_utils.tf2utils import MLP

from tf2_gnn.layers import WeightedSumGraphRepresentation, NodesToGraphRepresentationInput

from tf2_gnn.utils.gather_dense_gradient import gather_dense_gradient


class GraphGlobalExchangeInput(NamedTuple):
    """Input named tuple for graph global information exchange in GNNs."""

    node_embeddings: tf.Tensor
    node_to_graph_map: tf.Tensor
    num_graphs: tf.Tensor


class GraphGlobalExchange(tf.keras.layers.Layer):
    """Update node representations based on graph-global information."""

    def __init__(
        self,
        hidden_dim: int,
        weighting_fun: str = "softmax",
        num_heads: int = 4,
        dropout_rate: float = 0.0,
    ):
        """Initialise the layer."""
        super().__init__()
        self._hidden_dim = hidden_dim
        self._weighting_fun = weighting_fun
        self._num_heads = num_heads
        self._dropout_rate = dropout_rate

    def build(self, tensor_shapes: GraphGlobalExchangeInput):
        """Build the various layers in the model.

        Args:
            tensor_shapes: A GraphGlobalExchangeInput of tensor shapes.

        Returns:
            Nothing, but initialises the layers in the model based on the tensor shapes given.
        """
        self._node_to_graph_representation_layer = WeightedSumGraphRepresentation(
            graph_representation_size=self._hidden_dim,
            weighting_fun=self._weighting_fun,
            num_heads=self._num_heads,
            scoring_mlp_layers=[self._hidden_dim],
        )
        self._node_to_graph_representation_layer.build(
            NodesToGraphRepresentationInput(
                node_embeddings=tensor_shapes.node_embeddings,
                node_to_graph_map=tensor_shapes.node_to_graph_map,
                num_graphs=tensor_shapes.num_graphs,
            )
        )

        super().build(tensor_shapes)

    @abstractmethod
    def call(self, inputs: GraphGlobalExchangeInput, training: bool = False):
        """
        Args:
            inputs: A GraphGlobalExchangeInput containing the following fields:
                node_features: float32 tensor of shape [V, D], the original representation
                    of each node in the graph.
                
                node_to_graph_map: int32 tensor of shape [V], where node_to_graph_map[v] = i
                    means that node v belongs to graph i in the batch.
    
                num_graphs: int32 tensor of shape [], specifying number of graphs in batch.

            training: A bool representing whether the model is training or evaluating.

        Returns:
            A tensor of shape [V, hidden_dim]. The tensor represents the encoding of the
            states updated with information from the entire graph.
        """
        pass

    def _compute_per_node_graph_representations(
        self, inputs: GraphGlobalExchangeInput, training: bool = False
    ):
        cur_graph_representations = self._node_to_graph_representation_layer(
            NodesToGraphRepresentationInput(
                node_embeddings=inputs.node_embeddings,
                node_to_graph_map=inputs.node_to_graph_map,
                num_graphs=inputs.num_graphs,
            ),
            training=training,
        )  # Shape [G, hidden_dim]

        per_node_graph_representations = gather_dense_gradient(
            cur_graph_representations, inputs.node_to_graph_map
        )  # Shape [V, hidden_dim]

        if training:
            per_node_graph_representations = tf.nn.dropout(
                per_node_graph_representations, rate=self._dropout_rate
            )

        return per_node_graph_representations


class GraphGlobalMeanExchange(GraphGlobalExchange):
    def __init__(
        self,
        hidden_dim: int,
        weighting_fun: str = "softmax",
        num_heads: int = 4,
        dropout_rate: float = 0.0,
    ):
        """Initialise the layer."""
        super().__init__(hidden_dim, weighting_fun, num_heads, dropout_rate)

    def build(self, tensor_shapes: GraphGlobalExchangeInput):
        with tf.name_scope(self.__class__.__name__):
            super().build(tensor_shapes)

    def call(self, inputs: GraphGlobalExchangeInput, training: bool = False):
        per_node_graph_representations = self._compute_per_node_graph_representations(
            inputs, training
        )
        return (inputs.node_embeddings + per_node_graph_representations) / 2


class GraphGlobalGRUExchange(GraphGlobalExchange):
    def __init__(
        self,
        hidden_dim: int,
        weighting_fun: str = "softmax",
        num_heads: int = 4,
        dropout_rate: float = 0.0,
    ):
        """Initialise the layer."""
        super().__init__(hidden_dim, weighting_fun, num_heads, dropout_rate)

    def build(self, tensor_shapes: GraphGlobalExchangeInput):
        with tf.name_scope(self.__class__.__name__):
            self._gru_cell = tf.keras.layers.GRUCell(units=self._hidden_dim)
            self._gru_cell.build(tf.TensorShape((None, self._hidden_dim)))
            super().build(tensor_shapes)

    def call(self, inputs: GraphGlobalExchangeInput, training: bool = False):
        per_node_graph_representations = self._compute_per_node_graph_representations(
            inputs, training
        )
        cur_node_representations, _ = self._gru_cell(
            inputs=per_node_graph_representations,
            states=[inputs.node_embeddings],
            training=training,
        )
        return cur_node_representations


class GraphGlobalMLPExchange(GraphGlobalExchange):
    def __init__(
        self,
        hidden_dim: int,
        weighting_fun: str = "softmax",
        num_heads: int = 4,
        dropout_rate: float = 0.0,
    ):
        """Initialise the layer."""
        super().__init__(hidden_dim, weighting_fun, num_heads, dropout_rate)

    def build(self, tensor_shapes: GraphGlobalExchangeInput):
        with tf.name_scope(self.__class__.__name__):
            self._mlp = MLP(out_size=self._hidden_dim)
            self._mlp.build(tf.TensorShape((None, 2 * self._hidden_dim)))
            super().build(tensor_shapes)

    def call(self, inputs: GraphGlobalExchangeInput, training: bool = False):
        per_node_graph_representations = self._compute_per_node_graph_representations(
            inputs, training
        )
        cur_node_representations = self._mlp(
            tf.concat([per_node_graph_representations, inputs.node_embeddings], axis=-1),
            training=training,
        )
        return cur_node_representations



from typing import Any, Dict, Iterable, List, Iterator, Tuple, Optional, Set

import numpy as np
import tensorflow as tf
from dpu_utils.utils import RichPath

from .graph_dataset import DataFold, GraphSample, GraphBatchTFDataDescription, GraphDataset


class PPIGraphSample(GraphSample):
    """Data structure holding a single PPI graph."""

    def __init__(
        self,
        adjacency_lists: List[np.ndarray],
        type_to_node_to_num_inedges: np.ndarray,
        node_features: np.ndarray,
        node_labels: np.ndarray,
    ):
        super().__init__(adjacency_lists, type_to_node_to_num_inedges, node_features)
        self._node_labels = node_labels

    @property
    def node_labels(self) -> np.ndarray:
        """Node labels to predict as ndarray of shape [V, C]"""
        return self._node_labels


class PPIDataset(GraphDataset[PPIGraphSample]):
    @classmethod
    def get_default_hyperparameters(cls) -> Dict[str, Any]:
        return {
            "max_nodes_per_batch": 10000,
            "add_self_loop_edges": True,
            "tie_fwd_bkwd_edges": False,
        }

    @staticmethod
    def default_data_path() -> str:
        return "data/ppi"

    def __init__(self, params: Dict[str, Any], metadata: Optional[Dict[str, Any]] = None):
        super().__init__(params, metadata=metadata)

        self._num_edge_types = 1
        if params["add_self_loop_edges"]:
            self._num_edge_types += 1
        if not params["tie_fwd_bkwd_edges"]:
            self._num_edge_types += 1

        # Things that will be filled once we load data:
        self._loaded_data: Dict[DataFold, List[PPIGraphSample]] = {}

    @property
    def num_edge_types(self) -> int:
        return self._num_edge_types

    @property
    def node_feature_shape(self) -> Tuple:
        some_data_fold = next(iter(self._loaded_data.values()))
        return (some_data_fold[0].node_features.shape[-1],)

    @property
    def num_node_target_labels(self) -> int:
        return 121

    # -------------------- Data Loading --------------------
    def load_data(self, path: RichPath, folds_to_load: Optional[Set[DataFold]] = None) -> None:
        # Data in format as downloaded from https://s3.us-east-2.amazonaws.com/dgl.ai/dataset/ppi.zip
        # If we haven't defined what folds to load, load all:
        if folds_to_load is None:
            folds_to_load = {DataFold.TRAIN, DataFold.VALIDATION, DataFold.TEST}

        if DataFold.TRAIN in folds_to_load:
            self._loaded_data[DataFold.TRAIN] = self.__load_data(path, DataFold.TRAIN)
        if DataFold.VALIDATION in folds_to_load:
            self._loaded_data[DataFold.VALIDATION] = self.__load_data(path, DataFold.VALIDATION)
        if DataFold.TEST in folds_to_load:
            self._loaded_data[DataFold.TEST] = self.__load_data(path, DataFold.TEST)

    def load_data_from_list(
        self, datapoints: List[Dict[str, Any]], target_fold: DataFold = DataFold.TEST
    ):
        raise NotImplementedError()

    def __load_data(self, data_dir: RichPath, data_fold: DataFold) -> List[PPIGraphSample]:
        if data_fold == DataFold.TRAIN:
            data_name = "train"
        elif data_fold == DataFold.VALIDATION:
            data_name = "valid"
        elif data_fold == DataFold.TEST:
            data_name = "test"
        else:
            raise ValueError("Unknown data fold '%s'" % str(data_fold))
        print(" Loading PPI %s data from %s." % (data_name, data_dir))

        graph_json_data = data_dir.join("%s_graph.json" % data_name).read_by_file_suffix()
        node_to_features = data_dir.join("%s_feats.npy" % data_name).read_by_file_suffix()
        node_to_labels = data_dir.join("%s_labels.npy" % data_name).read_by_file_suffix()
        node_to_graph_id = data_dir.join("%s_graph_id.npy" % data_name).read_by_file_suffix()

        # We read in all the data in two steps:
        #  (1) Read features, labels and insert self-loop edges (edge type 0).
        #      Implicitly, this gives us the number of nodes per graph.
        #  (2) Read all edges, and shift them so that each graph starts with node 0.
        fwd_edge_type = 0
        if self._params["add_self_loop_edges"]:
            self_loop_edge_type = 1
        if not self._params["tie_fwd_bkwd_edges"]:
            bkwd_edge_type = self.num_edge_types - 1  # Always last

        graph_id_to_graph_data: Dict[int, PPIGraphSample] = {}
        graph_id_to_node_offset: Dict[int, int] = {}
        num_total_nodes = node_to_features.shape[0]
        for node_id in range(num_total_nodes):
            graph_id = node_to_graph_id[node_id]
            # In case we are entering a new graph, note its ID, so that we can normalise everything to start at 0
            if graph_id not in graph_id_to_graph_data:
                graph_id_to_graph_data[graph_id] = PPIGraphSample(
                    adjacency_lists=[[] for _ in range(self.num_edge_types)],
                    type_to_node_to_num_inedges=[[] for _ in range(self.num_edge_types)],
                    node_features=[],
                    node_labels=[],
                )
                graph_id_to_node_offset[graph_id] = node_id
            cur_graph_data = graph_id_to_graph_data[graph_id]
            cur_graph_data.node_features.append(node_to_features[node_id])
            cur_graph_data.node_labels.append(node_to_labels[node_id])
            shifted_node_id = node_id - graph_id_to_node_offset[graph_id]
            if self._params["add_self_loop_edges"]:
                cur_graph_data.adjacency_lists[self_loop_edge_type].append(
                    (shifted_node_id, shifted_node_id)
                )
                cur_graph_data.type_to_node_to_num_inedges[self_loop_edge_type].append(1)

        # Prepare reading of the edges by setting counters to 0:
        for graph_data in graph_id_to_graph_data.values():
            num_graph_nodes = len(graph_data.node_features)
            graph_data.type_to_node_to_num_inedges[fwd_edge_type] = np.zeros(
                [num_graph_nodes], np.int32
            )
            if not self._params["tie_fwd_bkwd_edges"]:
                graph_data.type_to_node_to_num_inedges[bkwd_edge_type] = np.zeros(
                    [num_graph_nodes], np.int32
                )

        for edge_info in graph_json_data["links"]:
            src_node, tgt_node = edge_info["source"], edge_info["target"]
            # First, shift node IDs so that each graph starts at node 0:
            graph_id = node_to_graph_id[src_node]
            graph_node_offset = graph_id_to_node_offset[graph_id]
            src_node, tgt_node = src_node - graph_node_offset, tgt_node - graph_node_offset

            cur_graph_data = graph_id_to_graph_data[graph_id]
            cur_graph_data.adjacency_lists[fwd_edge_type].append((src_node, tgt_node))
            cur_graph_data.type_to_node_to_num_inedges[fwd_edge_type][tgt_node] += 1
            if not self._params["tie_fwd_bkwd_edges"]:
                cur_graph_data.adjacency_lists[bkwd_edge_type].append((tgt_node, src_node))
                cur_graph_data.type_to_node_to_num_inedges[bkwd_edge_type][src_node] += 1
            else:
                cur_graph_data.adjacency_lists[fwd_edge_type].append((tgt_node, src_node))
                cur_graph_data.type_to_node_to_num_inedges[fwd_edge_type][src_node] += 1

        final_graphs = []
        for graph_data in graph_id_to_graph_data.values():
            # numpy-ize:
            adj_lists = []
            for edge_type_idx in range(self.num_edge_types):
                adj_lists.append(np.array(graph_data.adjacency_lists[edge_type_idx]))
            final_graphs.append(
                PPIGraphSample(
                    adjacency_lists=adj_lists,
                    type_to_node_to_num_inedges=np.array(graph_data.type_to_node_to_num_inedges),
                    node_features=np.array(graph_data.node_features),
                    node_labels=np.array(graph_data.node_labels),
                )
            )

        return final_graphs

    # -------------------- Minibatching --------------------
    def get_batch_tf_data_description(self) -> GraphBatchTFDataDescription:
        data_description = super().get_batch_tf_data_description()
        return GraphBatchTFDataDescription(
            batch_features_types=data_description.batch_features_types,
            batch_features_shapes=data_description.batch_features_shapes,
            batch_labels_types={**data_description.batch_labels_types, "node_labels": tf.float32},
            batch_labels_shapes={**data_description.batch_labels_shapes, "node_labels": (None, None)},
        )

    def _graph_iterator(self, data_fold: DataFold) -> Iterator[PPIGraphSample]:
        loaded_data = self._loaded_data[data_fold]
        if data_fold == DataFold.TRAIN:
            np.random.shuffle(loaded_data)
        return iter(loaded_data)

    def _new_batch(self) -> Dict[str, Any]:
        new_batch = super()._new_batch()
        new_batch["node_labels"] = []
        return new_batch

    def _add_graph_to_batch(self, raw_batch, graph_sample: PPIGraphSample) -> None:
        super()._add_graph_to_batch(raw_batch, graph_sample)
        raw_batch["node_labels"].append(graph_sample.node_labels)

    def _finalise_batch(self, raw_batch) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        batch_features, batch_labels = super()._finalise_batch(raw_batch)
        batch_labels["node_labels"] = np.concatenate(raw_batch["node_labels"], axis=0)
        return batch_features, batch_labels



from typing import Any, Dict, Iterable, List, NamedTuple, Iterator, Optional, Tuple

import numpy as np
import tensorflow as tf

from tf2_gnn.data import GraphDataset
from tf2_gnn.models import GraphTaskModel


def micro_f1(logits, labels):
    # Everything on int, because who trusts float anyway?
    predicted = tf.math.round(tf.nn.sigmoid(logits))
    predicted = tf.cast(predicted, dtype=tf.int32)
    labels = tf.cast(labels, dtype=tf.int32)

    true_pos = tf.math.count_nonzero(predicted * labels)
    false_pos = tf.math.count_nonzero(predicted * (labels - 1))
    false_neg = tf.math.count_nonzero((predicted - 1) * labels)

    precision = true_pos / (true_pos + false_pos)
    recall = true_pos / (true_pos + false_neg)
    fmeasure = (2 * precision * recall) / (precision + recall)
    return tf.cast(fmeasure, tf.float32)


class NodeMulticlassTask(GraphTaskModel):
    @classmethod
    def get_default_hyperparameters(cls, mp_style: Optional[str] = None) -> Dict[str, Any]:
        super_params = super().get_default_hyperparameters(mp_style)
        these_hypers: Dict[str, Any] = {}
        super_params.update(these_hypers)
        return super_params

    def __init__(self, params: Dict[str, Any], dataset: GraphDataset, name: str = None):
        super().__init__(params, dataset=dataset, name=name)
        if not hasattr(dataset, "num_node_target_labels"):
            raise ValueError(f"Provided dataset of type {type(dataset)} does not provide num_node_target_labels information.")
        self._num_labels = dataset.num_node_target_labels

    def build(self, input_shapes):
        with tf.name_scope(self.__class__.__name__):
            self.node_to_labels_layer = tf.keras.layers.Dense(units=self._num_labels, use_bias=True)
            self.node_to_labels_layer.build((None, self._params["gnn_hidden_dim"]))
        super().build(input_shapes)

    def compute_task_output(
        self, batch_features, final_node_representations: tf.Tensor, training: bool
    ):
        per_node_logits = self.node_to_labels_layer(final_node_representations)
        return (per_node_logits,)

    def compute_task_metrics(
        self, batch_features, task_output, batch_labels
    ) -> Dict[str, tf.Tensor]:
        (per_node_logits,) = task_output
        (loss, f1_score) = self._fast_task_metrics(per_node_logits, batch_labels["node_labels"])

        return {"loss": loss, "f1_score": f1_score}

    @tf.function(input_signature=(tf.TensorSpec((None, None)), tf.TensorSpec((None, None))))
    def _fast_task_metrics(self, per_node_logits, node_labels):
        per_node_losses = tf.nn.sigmoid_cross_entropy_with_logits(
            logits=per_node_logits, labels=node_labels
        )
        loss = tf.reduce_mean(tf.reduce_sum(per_node_losses, axis=-1))  # Compute mean loss _per node_
        f1_score = micro_f1(per_node_logits, node_labels)

        return loss, f1_score

    def compute_epoch_metrics(self, task_results: List[Any]) -> Tuple[float, str]:
        avg_microf1 = np.average([r["f1_score"] for r in task_results])
        return -avg_microf1, f"Avg MicroF1: {avg_microf1:.3f}"



"""General dataset class for datasets with a numeric property stored as JSONLines files."""
from typing import Any, Dict, List, Optional, Tuple, TypeVar

import numpy as np
import tensorflow as tf

from .graph_dataset import GraphBatchTFDataDescription, GraphSample
from .jsonl_graph_dataset import JsonLGraphDataset


class GraphWithPropertySample(GraphSample):
    """Data structure holding a single graph with a single numeric property."""

    def __init__(
        self,
        adjacency_lists: List[np.ndarray],
        type_to_node_to_num_incoming_edges: np.ndarray,
        node_features: List[np.ndarray],
        target_value: float,
    ):
        super().__init__(adjacency_lists, type_to_node_to_num_incoming_edges, node_features)
        self._target_value = target_value

    @property
    def target_value(self) -> float:
        """Target value of the regression task."""
        return self._target_value

    def __str__(self):
        return (
            f"Adj:            {self._adjacency_lists}\n"
            f"Node_features:  {self._node_features}\n"
            f"Target_value:   {self._target_value}"
        )


GraphWithPropertySampleType = TypeVar(
    "GraphWithPropertySampleType", bound=GraphWithPropertySample
)


class JsonLGraphPropertyDataset(JsonLGraphDataset[GraphWithPropertySampleType]):
    """
    General class representing pre-split datasets in JSONLines format.
    Concretely, this class expects the following:
    * In the data directory, files "train.jsonl.gz", "valid.jsonl.gz" and
      "test.jsonl.gz" are used to store the train/valid/test datasets.
    * Each of the files is gzipped text file in which each line is a valid
      JSON dictionary with a "graph" key, which in turn points to a
      dictionary with keys
       - "node_features" (list of numerical initial node labels)
       - "adjacency_lists" (list of list of directed edge pairs)
      Addtionally, the dictionary has to contain a "Property" key with a
      floating point value.
    """

    @classmethod
    def get_default_hyperparameters(cls) -> Dict[str, Any]:
        super_hypers = super().get_default_hyperparameters()
        this_hypers = {
            # If None, the data-provided property is used; otherwise, a floating point
            # value is expected and property values greater than this value will be
            # encoded as 1.0 and smaller values will be encoded as 0.0.
            "threshold_for_classification": None,
        }
        super_hypers.update(this_hypers)
        return super_hypers

    def __init__(
        self, params: Dict[str, Any], metadata: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(params, metadata=metadata)
        self._threshold_for_classification = params["threshold_for_classification"]

    def _process_raw_datapoint(
        self, datapoint: Dict[str, Any]
    ) -> GraphWithPropertySampleType:
        node_features = datapoint["graph"]["node_features"]
        type_to_adj_list, type_to_num_incoming_edges = self._process_raw_adjacency_lists(
            raw_adjacency_lists=datapoint["graph"]["adjacency_lists"],
            num_nodes=len(node_features),
        )

        target_value = float(datapoint["Property"])
        if self._threshold_for_classification is not None:
            target_value = float(target_value > self._threshold_for_classification)

        return GraphWithPropertySample(
            adjacency_lists=type_to_adj_list,
            type_to_node_to_num_incoming_edges=type_to_num_incoming_edges,
            node_features=node_features,
            target_value=target_value,
        )

    def _new_batch(self) -> Dict[str, Any]:
        new_batch = super()._new_batch()
        new_batch["target_value"] = []
        return new_batch

    def _add_graph_to_batch(
        self, raw_batch: Dict[str, Any], graph_sample: GraphWithPropertySampleType
    ) -> None:
        super()._add_graph_to_batch(raw_batch, graph_sample)
        raw_batch["target_value"].append(graph_sample.target_value)

    def _finalise_batch(self, raw_batch) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        batch_features, batch_labels = super()._finalise_batch(raw_batch)
        return batch_features, {"target_value": raw_batch["target_value"]}

    def get_batch_tf_data_description(self) -> GraphBatchTFDataDescription:
        data_description = super().get_batch_tf_data_description()
        return GraphBatchTFDataDescription(
            batch_features_types=data_description.batch_features_types,
            batch_features_shapes=data_description.batch_features_shapes,
            batch_labels_types={**data_description.batch_labels_types, "target_value": tf.float32},
            batch_labels_shapes={**data_description.batch_labels_shapes, "target_value": (None,)},
        )
